import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from typing import Optional
import matplotlib.figure

class CitrosStat:
    """
    Object to store statistics.

    Parameters
    ----------
    F : pandas.DataFrame
        Table with statistics.
    labels : array-like
        Labels of the data columns.
    x_label : str
        Label of the independent variable.

    Attributes
    ----------
    x : pandas.DataFrame
        Table with independent variable.
    mean : pandas.DataFrame
        Table with mean values. If statistics was collected for a vector, columns correspond to vector elements.
    covar_matrix: pandas.DataFrame
        Table with the covariance matrixes. If statistics was collected for a vector, columns correspond to vector elements.
    sigma : pandas.DataFrame
        Table with the square roots of the diagonal elements of the covariance matrix. 
        If statistics was collected for a vector, columns correspond to vector elements.
    """

    def __init__(self, F, labels, x_label):
        self.x = pd.DataFrame(F[x_label])
        self.mean = pd.DataFrame(columns = labels)
        self.covar_matrix = F['covar_matrix']
        self.sigma = self.std = pd.DataFrame(columns = labels)
        for i in range(len(labels)):
            self.mean[labels[i]] = F['mean'].apply(lambda x: x[i])
            self.sigma[labels[i]] = F['sigma'].apply(lambda x: x[i])

    def to_pandas(self):
        """
        Convert CitrosStat object back to pandas DataFrame.

        Returns
        -------
        df : pandas.DataFrame
            Converted to pandas DataFrame.
        """
        result = pd.concat([self.x, self.mean.apply(lambda x: np.array([x[col] for col in self.mean.columns]), axis = 1),
            self.covar_matrix, self.sigma.apply(lambda x: np.array([x[col] for col in self.mean.columns]), axis = 1)], 
            axis = 1)
        result.columns = [result.columns[0], 'mean', 'covar_matrix', 'sigma']
        return result
    
    def plot(self, fig: Optional[matplotlib.figure.Figure] = None, show_fig: bool = True, return_fig: bool = False, 
             n_std: int = 3, fig_title: str = 'Statistics', std_color: str = 'r'):
        """
        Plot mean values and standard deviations.

        Parameters
        ----------
        fig : matplotlib.figure.Figure
            figure to plot on. If None, the new one will be created.
        show_fig : bool
            If the figure should be shown, True by default.
        return_fig : bool
            If the figure parameters fig, ax should be returned; 
            fig is matplotlib.figure.Figure and ax is matplotlib.axes.Axes
        n_std : int, default 3
            Error interval to display, specified in standard deviations.
        fig_title : str, default 'Statistics'
            Title of the figure.
        std_color : str, default 'r'
            Color for displaying standard deviations, red by default.

        Returns
        -------
        fig : matplotlib.figure.Figure
            if `return_fig` set to True
        ax : list of matplotlib.axes.Axes
            if `return_fig` set to True
        """
        if (not show_fig) and (not return_fig):
            return

        N = len(self.mean.columns)

        if fig is None:
            fig, axes = plt.subplots(nrows = N, ncols = 1,figsize=(6, 6))
            if N == 1:
                axes = [axes]
        else:
            axes = []
            for i in range(1,N+1):
                axes.append(fig.add_subplot(N,1,i))

        for i, ax in enumerate(axes):
            filter_st = self.sigma.iloc[:, i].notna()

            ax.plot(self.x, self.mean.iloc[:, i], 'k-', linewidth = 2, label = 'mean')
            if n_std is not None:
                ax.plot(self.x[filter_st], self.mean.iloc[:, i][filter_st] + n_std*self.sigma.iloc[:, i][filter_st],
                        '-', color = std_color, label = r'$\pm$'+str(n_std)+ r'$\sigma$')
                ax.plot(self.x[filter_st], self.mean.iloc[:, i][filter_st] - n_std*self.sigma.iloc[:, i][filter_st], 
                        '-', color = std_color)
            ax.grid(True)
            ax.set_ylabel(self.mean.columns[i])
        handles, labels = axes[-1].get_legend_handles_labels()

        fig.suptitle(fig_title)
        fig.supylabel(self.x.columns[0])
        fig.legend(handles, labels, bbox_to_anchor=(1.0, 0.94),loc ='upper left')
        fig.tight_layout()

        if show_fig:
            fig.show()
        if return_fig:
            return fig, ax


import os
import uvicorn
import logging
from citros import get_logger, shutdown_log
from pathlib import Path
from fastapi import Depends, FastAPI, HTTPException, BackgroundTasks
from citros.batch import Batch, NoBatchFoundException
from fastapi_utils.timing import add_timing_middleware, record_timing

app = FastAPI()


class NoDataFoundException(Exception):
    def __init__(self, message="No Data found."):
        super().__init__(message)


# get batch info
@app.get("/{simulation}/{batch_name}")
async def get_batch(simulation, batch_name):
    try:
        batch = Batch(
            app.root,
            simulation,
            batch_name,
            log=app.log,
            debug=app.debug,
            verbose=app.verbose,
        )
        return batch.data
    except NoBatchFoundException:
        raise HTTPException(status_code=404, detail="Item not found")


def load_batch_run(simulation: str, batch_name: str, message=""):
    # TODO[enhancement]: service worker that will load the data to DB
    pass


# request access to batch, loads the batch simulations to DB
@app.post("/{simulation}/{batch_name}")
async def request_access_batch(
    simulation: str, batch_name: str, background_tasks: BackgroundTasks
):
    background_tasks.add_task(
        load_batch_run, simulation, batch_name, message="some notification"
    )
    return {"message": "Batch run {simulation}/{batch_name} is loading. please wait"}


def data_access_service(
    root, time=False, host="localhost", port=8080, debug=False, verbose=False
):
    app.root = root
    if not root:
        raise ValueError("root path is required")
    if not Path(root).exists():
        raise NoDataFoundException(f"root path {root} does not exists")
    app.debug = debug
    app.verbose = verbose
    app.log = get_logger(
        __name__,
        log_level=os.environ.get("LOGLEVEL", "DEBUG" if debug else "INFO"),
        log_file=str(Path(root) / "citros.log"),
        verbose=verbose,
    )

    loglevel = logging._nameToLevel["ERROR"]
    # if debug:
    #     loglevel = logging._nameToLevel["DEBUG"]
    # logging.basicConfig(level=logging.DEBUG)
    if verbose:
        loglevel = logging._nameToLevel["INFO"]
        # logging.basicConfig(level=logging.INFO)
    # logging.basicConfig(level=logging.ERROR)  # second call do nothing...

    if time:
        logger = logging.getLogger("citros.data_access")
        add_timing_middleware(app, record=logger.info, prefix="app", exclude="untimed")

    from fastapi.logger import logger as fastapi_logger

    # app.log.
    uvicorn.run(app, host=host, port=int(port), log_level=loglevel)



import argparse
import importlib_resources
from bin.cli_impl import *
from rich import print, inspect, print_json
from rich.rule import Rule
from rich.panel import Panel
from rich.padding import Padding
from rich.logging import RichHandler
from rich.console import Console
from rich.markdown import Markdown
from rich.traceback import install
from .formatter import RichHelpFormatterCitros

install()
from bin import __version__ as citros_version


def parser_run(main_sub, epilog=None):
    parser = main_sub.add_parser(
        "run",
        description=Panel(
            Markdown(
                open(
                    importlib_resources.files(f"data.doc.cli").joinpath("run.md"), "r"
                ).read()
            ),
        ),
        epilog=epilog,
        help="Run ROS simulation using CITROS",
        formatter_class=RichHelpFormatterCitros,
    )
    parser.add_argument("-dir", default=".", help="The working dir of the project")
    parser.add_argument(
        "-s", "--simulation", nargs="?", default=None, help="Simulation name"
    )
    parser.add_argument(
        "-n", "--name", nargs="?", default=None, help="Name for the batch run"
    )
    parser.add_argument(
        "-m", "--message", nargs="?", default=None, help="Message for the batch run"
    )
    parser.add_argument(
        "--version",
        nargs="?",
        default=None,
        help="the batch run version. .citros/data/{sim_name}/{batch_name}/{ **batch_version** }/{sid}, if not provided will default to the current date",
    )

    parser.add_argument(
        "-l",
        "--lan_traffic",
        action="store_true",
        help="receive LAN ROS traffic in your simulation.",
    )
    parser.add_argument(
        "-c",
        "--completions",
        nargs="?",
        default=1,
        help="number of times to run the simulation",
    )
    parser.add_argument(
        "-i",
        "--index",
        nargs="?",
        default=-1,
        help="the pard of the batch. if -1 will run all the batch",
    )

    parser.add_argument(
        "-d", "--debug", action="store_true", help="set logging level to debug"
    )
    parser.add_argument(
        "-v", "--verbose", action="store_true", help="use verbose console prints"
    )
    parser.set_defaults(func=run)



import argparse
import importlib_resources
from bin.cli_impl import *
from rich import print, inspect, print_json
from rich.rule import Rule
from rich.panel import Panel
from rich.padding import Padding
from rich.logging import RichHandler
from rich.console import Console
from rich.markdown import Markdown
from rich.traceback import install
from bin import __version__ as citros_version
from .formatter import RichHelpFormatterCitros

install()


# citros data list
def parser_data_list(parent_subparser, epilog=None):
    description_path = "data/list.md"
    help = "Will list all recorded batches under .citros/data folder as a table"

    parser = parent_subparser.add_parser(
        "list",
        description=Panel(
            Markdown(
                open(
                    importlib_resources.files(f"data.doc.cli").joinpath(
                        description_path
                    ),
                    "r",
                ).read()
            ),
            subtitle=f"[{citros_version}]",
            title="description",
        ),
        epilog=epilog,
        help=help,
        formatter_class=RichHelpFormatterCitros,
    )
    parser.add_argument(
        "-dir", "--dir", default=".", help="The working dir of the project"
    )
    parser.add_argument(
        "-d", "--debug", action="store_true", help="set logging level to debug"
    )
    parser.add_argument(
        "-v", "--verbose", action="store_true", help="use verbose console prints"
    )
    parser.set_defaults(func=data_list)

    return parser


def parser_data_tree(parent_subparser, epilog=None):
    description_path = "data/tree.md"
    help = "Will list all recorded batches under .citros/data folder in a tree view"

    parser = parent_subparser.add_parser(
        "tree",
        description=Panel(
            Markdown(
                open(
                    importlib_resources.files(f"data.doc.cli").joinpath(
                        description_path
                    ),
                    "r",
                ).read()
            ),
            subtitle=f"[{citros_version}]",
            title="description",
        ),
        epilog=epilog,
        help=help,
        formatter_class=RichHelpFormatterCitros,
    )
    parser.add_argument(
        "-dir", "--dir", default=".", help="The working dir of the project"
    )
    parser.add_argument(
        "-d", "--debug", action="store_true", help="set logging level to debug"
    )
    parser.add_argument(
        "-v", "--verbose", action="store_true", help="use verbose console prints"
    )
    parser.set_defaults(func=data_tree)

    subsubparser = parser.add_subparsers(dest="type")
    parser_data_tree_info(subsubparser, epilog)
    parser_data_tree_load(subsubparser, epilog)
    parser_data_tree_unload(subsubparser, epilog)
    parser_data_tree_delete(subsubparser, epilog)

    return parser


# citros data tree load
def parser_data_tree_info(parent_subparser, epilog=None):
    description_path = "data/tree/info.md"
    help = "will show batch info "

    parser = parent_subparser.add_parser(
        "info",
        description=Panel(
            Markdown(
                open(
                    importlib_resources.files(f"data.doc.cli").joinpath(
                        description_path
                    ),
                    "r",
                ).read()
            ),
            subtitle=f"[{citros_version}]",
            title="description",
        ),
        epilog=epilog,
        help=help,
        formatter_class=RichHelpFormatterCitros,
    )
    parser.add_argument("-s", "--simulation", help="simulation name")
    parser.add_argument("-b", "--batch", default="citros", help="batch name")
    parser.add_argument("--version", help="batch version ")

    parser.add_argument(
        "-d", "--debug", action="store_true", help="set logging level to debug"
    )
    parser.add_argument(
        "-v", "--verbose", action="store_true", help="use verbose console prints"
    )
    parser.set_defaults(func=data_info)

    return parser


# citros data tree load
def parser_data_tree_load(parent_subparser, epilog=None):
    description_path = "data/tree/load.md"
    help = "will load the batch to db"

    parser = parent_subparser.add_parser(
        "load",
        description=Panel(
            Markdown(
                open(
                    importlib_resources.files(f"data.doc.cli").joinpath(
                        description_path
                    ),
                    "r",
                ).read()
            ),
            subtitle=f"[{citros_version}]",
            title="description",
        ),
        epilog=epilog,
        help=help,
        formatter_class=RichHelpFormatterCitros,
    )
    parser.add_argument("-s", "--simulation", help="simulation name")
    parser.add_argument("-b", "--batch", default="citros", help="batch name")
    parser.add_argument("--version", help="batch version ")
    parser.add_argument(
        "-d", "--debug", action="store_true", help="set logging level to debug"
    )
    parser.add_argument(
        "-v", "--verbose", action="store_true", help="use verbose console prints"
    )
    parser.set_defaults(func=data_load)

    return parser


# citros data tree load
def parser_data_tree_unload(parent_subparser, epilog=None):
    description_path = "data/tree/unload.md"
    help = "will unload the batch from db"

    parser = parent_subparser.add_parser(
        "unload",
        description=Panel(
            Markdown(
                open(
                    importlib_resources.files(f"data.doc.cli").joinpath(
                        description_path
                    ),
                    "r",
                ).read()
            ),
            subtitle=f"[{citros_version}]",
            title="description",
        ),
        epilog=epilog,
        help=help,
        formatter_class=RichHelpFormatterCitros,
    )
    parser.add_argument("-s", "--simulation", help="simulation name")
    parser.add_argument("-b", "--batch", default="citros", help="batch name")
    parser.add_argument("--version", help="batch version ")
    parser.add_argument(
        "-d", "--debug", action="store_true", help="set logging level to debug"
    )
    parser.add_argument(
        "-v", "--verbose", action="store_true", help="use verbose console prints"
    )
    parser.set_defaults(func=data_unload)

    return parser


# citros data tree load
def parser_data_tree_delete(parent_subparser, epilog=None):
    description_path = "data/tree/delete.md"
    help = "will delete the batch"

    parser = parent_subparser.add_parser(
        "delete",
        description=Panel(
            Markdown(
                open(
                    importlib_resources.files(f"data.doc.cli").joinpath(
                        description_path
                    ),
                    "r",
                ).read()
            ),
            subtitle=f"[{citros_version}]",
            title="description",
        ),
        epilog=epilog,
        help=help,
        formatter_class=RichHelpFormatterCitros,
    )
    parser.add_argument("-s", "--simulation", help="simulation name")
    parser.add_argument("-b", "--batch", default="citros", help="batch name")
    parser.add_argument("--version", help="batch version ")
    parser.add_argument(
        "-d", "--debug", action="store_true", help="set logging level to debug"
    )
    parser.add_argument(
        "-v", "--verbose", action="store_true", help="use verbose console prints"
    )
    parser.set_defaults(func=data_delete)

    return parser


# citros data db create
def parser_data_db_create(parent_subparser, epilog=None):
    description_path = "data/db/create.md"
    help = "Create and init new DB docker instance"

    parser = parent_subparser.add_parser(
        "create",
        description=Panel(
            Markdown(
                open(
                    importlib_resources.files(f"data.doc.cli").joinpath(
                        description_path
                    ),
                    "r",
                ).read()
            ),
            subtitle=f"[{citros_version}]",
            title="description",
        ),
        epilog=epilog,
        help=help,
        formatter_class=RichHelpFormatterCitros,
    )

    parser.add_argument(
        "-d", "--debug", action="store_true", help="set logging level to debug"
    )
    parser.add_argument(
        "-v", "--verbose", action="store_true", help="use verbose console prints"
    )
    parser.set_defaults(func=data_db_create)

    # subparser = parser.add_subparsers(dest="type")

    return parser


# citros data db init
def parser_data_db_init(parent_subparser, epilog=None):
    description_path = "data/db/init.md"
    help = "initialize DB for use with citros."

    parser = parent_subparser.add_parser(
        "init",
        description=Panel(
            Markdown(
                open(
                    importlib_resources.files(f"data.doc.cli").joinpath(
                        description_path
                    ),
                    "r",
                ).read()
            ),
            subtitle=f"[{citros_version}]",
            title="description",
        ),
        epilog=epilog,
        help=help,
        formatter_class=RichHelpFormatterCitros,
    )

    parser.add_argument(
        "-d", "--debug", action="store_true", help="set logging level to debug"
    )
    parser.add_argument(
        "-v", "--verbose", action="store_true", help="use verbose console prints"
    )
    parser.set_defaults(func=data_db_init)

    # subparser = parser.add_subparsers(dest="type")

    return parser


# citros data db status
def parser_data_db_status(parent_subparser, epilog=None):
    description_path = "data/db/status.md"
    help = "Get DB status"

    parser = parent_subparser.add_parser(
        "status",
        description=Panel(
            Markdown(
                open(
                    importlib_resources.files(f"data.doc.cli").joinpath(
                        description_path
                    ),
                    "r",
                ).read()
            ),
            subtitle=f"[{citros_version}]",
            title="description",
        ),
        epilog=epilog,
        help=help,
        formatter_class=RichHelpFormatterCitros,
    )

    parser.add_argument(
        "-d", "--debug", action="store_true", help="set logging level to debug"
    )
    parser.add_argument(
        "-v", "--verbose", action="store_true", help="use verbose console prints"
    )
    parser.set_defaults(func=data_db_status)

    # subparser = parser.add_subparsers(dest="type")

    return parser


# citros data db stop
def parser_data_db_stop(parent_subparser, epilog=None):
    description_path = "data/db/stop.md"
    help = "Stop the DB"

    parser = parent_subparser.add_parser(
        "stop",
        description=Panel(
            Markdown(
                open(
                    importlib_resources.files(f"data.doc.cli").joinpath(
                        description_path
                    ),
                    "r",
                ).read()
            ),
            subtitle=f"[{citros_version}]",
            title="description",
        ),
        epilog=epilog,
        help=help,
        formatter_class=RichHelpFormatterCitros,
    )

    parser.add_argument(
        "-d", "--debug", action="store_true", help="set logging level to debug"
    )
    parser.add_argument(
        "-v", "--verbose", action="store_true", help="use verbose console prints"
    )
    parser.set_defaults(func=data_db_stop)

    # subparser = parser.add_subparsers(dest="type")

    return parser


# citros data db start
def parser_data_db_start(parent_subparser, epilog=None):
    description_path = "data/db/start.md"
    help = "Start the DB"

    parser = parent_subparser.add_parser(
        "start",
        description=Panel(
            Markdown(
                open(
                    importlib_resources.files(f"data.doc.cli").joinpath(
                        description_path
                    ),
                    "r",
                ).read()
            ),
            subtitle=f"[{citros_version}]",
            title="description",
        ),
        epilog=epilog,
        help=help,
        formatter_class=RichHelpFormatterCitros,
    )

    parser.add_argument(
        "-d", "--debug", action="store_true", help="set logging level to debug"
    )
    parser.add_argument(
        "-v", "--verbose", action="store_true", help="use verbose console prints"
    )
    parser.set_defaults(func=data_db_start)

    # subparser = parser.add_subparsers(dest="type")

    return parser


# citros data db logs
def parser_data_db_logs(parent_subparser, epilog=None):
    description_path = "data/db/logs.md"
    help = "Get DB logs"

    parser = parent_subparser.add_parser(
        "logs",
        description=Panel(
            Markdown(
                open(
                    importlib_resources.files(f"data.doc.cli").joinpath(
                        description_path
                    ),
                    "r",
                ).read()
            ),
            subtitle=f"[{citros_version}]",
            title="description",
        ),
        epilog=epilog,
        help=help,
        formatter_class=RichHelpFormatterCitros,
    )

    parser.add_argument(
        "-d", "--debug", action="store_true", help="set logging level to debug"
    )
    parser.add_argument(
        "-v", "--verbose", action="store_true", help="use verbose console prints"
    )
    parser.set_defaults(func=data_db_logs)

    # subparser = parser.add_subparsers(dest="type")

    return parser


# citros data db clean
def parser_data_db_clean(parent_subparser, epilog=None):
    description_path = "data/db/clean.md"
    help = "Clean all data from DB"

    parser = parent_subparser.add_parser(
        "clean",
        description=Panel(
            Markdown(
                open(
                    importlib_resources.files(f"data.doc.cli").joinpath(
                        description_path
                    ),
                    "r",
                ).read()
            ),
            subtitle=f"[{citros_version}]",
            title="description",
        ),
        epilog=epilog,
        help=help,
        formatter_class=RichHelpFormatterCitros,
    )

    parser.add_argument(
        "-d", "--debug", action="store_true", help="set logging level to debug"
    )
    parser.add_argument(
        "-v", "--verbose", action="store_true", help="use verbose console prints"
    )
    parser.set_defaults(func=data_db_clean)

    # subparser = parser.add_subparsers(dest="type")

    return parser


def parser_data_db_remove(parent_subparser, epilog=None):
    description_path = "data/db/remove.md"
    help = "Remove DB instance"

    parser = parent_subparser.add_parser(
        "remove",
        description=Panel(
            Markdown(
                open(
                    importlib_resources.files(f"data.doc.cli").joinpath(
                        description_path
                    ),
                    "r",
                ).read()
            ),
            subtitle=f"[{citros_version}]",
            title="description",
        ),
        epilog=epilog,
        help=help,
        formatter_class=RichHelpFormatterCitros,
    )

    parser.add_argument(
        "-d", "--debug", action="store_true", help="set logging level to debug"
    )
    parser.add_argument(
        "-v", "--verbose", action="store_true", help="use verbose console prints"
    )
    parser.set_defaults(func=data_db_remove)

    # subparser = parser.add_subparsers(dest="type")

    return parser


# citros data db
def parser_data_db(parent_subparser, epilog=None):
    description_path = "data/db.md"
    help = "data db section"

    parser = parent_subparser.add_parser(
        "db",
        description=Panel(
            Markdown(
                open(
                    importlib_resources.files(f"data.doc.cli").joinpath(
                        description_path
                    ),
                    "r",
                ).read()
            ),
            subtitle=f"[{citros_version}]",
            title="description",
        ),
        epilog=epilog,
        help=help,
        formatter_class=RichHelpFormatterCitros,
    )
    parser.add_argument(
        "-d", "--debug", action="store_true", help="set logging level to debug"
    )
    parser.add_argument(
        "-v", "--verbose", action="store_true", help="use verbose console prints"
    )
    parser.set_defaults(func=data_db)

    subparser = parser.add_subparsers(dest="type")

    parser_data_db_create(subparser)
    parser_data_db_remove(subparser)

    parser_data_db_init(subparser)
    parser_data_db_clean(subparser)

    parser_data_db_status(subparser)
    parser_data_db_start(subparser)
    parser_data_db_stop(subparser)
    parser_data_db_logs(subparser)

    return parser


# citros data
def parser_data(main_sub, epilog=None):
    description_path = "data.md"
    help = "Data related functionality"

    parser = main_sub.add_parser(
        "data",
        description=Panel(
            Markdown(
                open(
                    importlib_resources.files(f"data.doc.cli").joinpath(
                        description_path
                    ),
                    "r",
                ).read()
            ),
            subtitle=f"[{citros_version}]",
            title="description",
        ),
        epilog=epilog,
        help=help,
        formatter_class=RichHelpFormatterCitros,
    )
    parser.add_argument("-dir", default=".", help="The working dir of the project")
    parser.add_argument(
        "-d", "--debug", action="store_true", help="set logging level to debug"
    )
    parser.add_argument(
        "-v", "--verbose", action="store_true", help="use verbose console prints"
    )
    parser.set_defaults(func=data)

    subsubparser = parser.add_subparsers(dest="type")
    parser_data_list(subsubparser, epilog)
    parser_data_tree(subsubparser, epilog)
    parser_data_db(subsubparser, epilog)

    return parser



import argparse
import importlib_resources
from bin.cli_impl import *
from rich import print, inspect, print_json
from rich.rule import Rule
from rich.panel import Panel
from rich.padding import Padding
from rich.logging import RichHandler
from rich.console import Console
from rich.markdown import Markdown
from rich.traceback import install
from bin import __version__ as citros_version
from .formatter import RichHelpFormatterCitros

install()


def parser_data_start(parent_subparser, epilog=None):
    description_path = "service/start.md"
    help = "service start"

    parser = parent_subparser.add_parser(
        "start",
        description=Panel(
            Markdown(
                open(
                    importlib_resources.files(f"data.doc.cli").joinpath(
                        description_path
                    ),
                    "r",
                ).read()
            ),
            subtitle=f"[{citros_version}]",
            title="description",
        ),
        epilog=epilog,
        help=help,
        formatter_class=RichHelpFormatterCitros,
    )

    parser.add_argument("-dir", default=".", help="The working dir of the project")
    parser.add_argument("-H", "--host", default="0.0.0.0", help="host")
    parser.add_argument("-p", "--port", default="8000", help="post to listen to")
    parser.add_argument("-t", "--time", action="store_true", help="print request times")
    parser.add_argument(
        "-d", "--debug", action="store_true", help="set logging level to debug"
    )
    parser.add_argument(
        "-v", "--verbose", action="store_true", help="use verbose console prints"
    )
    parser.set_defaults(func=service_start)

    return parser


def parser_data_stop(parent_subparser, epilog=None):
    description_path = "service/stop.md"
    help = "service stop"

    parser = parent_subparser.add_parser(
        "stop",
        description=Panel(
            Markdown(
                open(
                    importlib_resources.files(f"data.doc.cli").joinpath(
                        description_path
                    ),
                    "r",
                ).read()
            ),
            subtitle=f"[{citros_version}]",
            title="description",
        ),
        epilog=epilog,
        help=help,
        formatter_class=RichHelpFormatterCitros,
    )
    parser.add_argument(
        "-dir", "--dir", default=".", help="The working dir of the project"
    )
    parser.add_argument(
        "-d", "--debug", action="store_true", help="set logging level to debug"
    )
    parser.add_argument(
        "-v", "--verbose", action="store_true", help="use verbose console prints"
    )
    parser.set_defaults(func=service_stop)

    return parser


def parser_data_status(parent_subparser, epilog=None):
    description_path = "service/status.md"
    help = "service status"

    parser = parent_subparser.add_parser(
        "status",
        description=Panel(
            Markdown(
                open(
                    importlib_resources.files(f"data.doc.cli").joinpath(
                        description_path
                    ),
                    "r",
                ).read()
            ),
            subtitle=f"[{citros_version}]",
            title="description",
        ),
        epilog=epilog,
        help=help,
        formatter_class=RichHelpFormatterCitros,
    )
    parser.add_argument(
        "-dir", "--dir", default=".", help="The working dir of the project"
    )
    parser.add_argument(
        "-d", "--debug", action="store_true", help="set logging level to debug"
    )
    parser.add_argument(
        "-v", "--verbose", action="store_true", help="use verbose console prints"
    )
    parser.set_defaults(func=service_status)

    return parser


# citros data
def parser_service(main_sub, epilog=None):
    description_path = "service.md"
    help = "service section"

    parser = main_sub.add_parser(
        "service",
        description=Panel(
            Markdown(
                open(
                    importlib_resources.files(f"data.doc.cli").joinpath(
                        description_path
                    ),
                    "r",
                ).read()
            ),
            subtitle=f"[{citros_version}]",
            title="description",
        ),
        epilog=epilog,
        help=help,
        formatter_class=RichHelpFormatterCitros,
    )
    parser.add_argument("-dir", default=".", help="The working dir of the project")
    parser.add_argument(
        "-d", "--debug", action="store_true", help="set logging level to debug"
    )
    parser.add_argument(
        "-v", "--verbose", action="store_true", help="use verbose console prints"
    )
    parser.set_defaults(func=service)

    subsubparser = parser.add_subparsers(dest="type")
    parser_data_start(subsubparser, epilog)
    parser_data_stop(subsubparser, epilog)
    parser_data_status(subsubparser, epilog)

    return parser



import argparse
import importlib_resources
from bin.cli_impl import *
from rich import print, inspect, print_json
from rich.rule import Rule
from rich.panel import Panel
from rich.padding import Padding
from rich.logging import RichHandler
from rich.console import Console
from rich.markdown import Markdown
from rich.traceback import install
from .parsers.formatter import RichHelpFormatterCitros

install()

from bin import __version__ as citros_version

from .parsers import (
    parser_run,
    parser_init,
    parser_data,
    parser_service,
    parser_report,
    parser_launch,
    parser_parameter,
    parser_simulation,
)


# PANNEL = ""
PANNEL = Panel.fit(
    f"""[green]████████████████████████████████████████████████████████████
██      ███        ██        ██       ████      ████      ██
█  ████  █████  ████████  █████  ████  ██  ████  ██  ███████
█  ███████████  ████████  █████       ███  ████  ███      ██
█  ████  █████  ████████  █████  ███  ███  ████  ████████  █
██      ███        █████  █████  ████  ███      ████      ██
████████████████████████████████████████████████████████████""",
    subtitle=f"[{citros_version}]",
)


def epilog(url="https://docs.citros.io"):
    return Panel(Markdown(f"Read more at `{url}`"))


def main():
    description_path = "citros.md"

    # main parser -----------------------------------------
    parser = argparse.ArgumentParser(
        description=Panel(
            Markdown(
                open(
                    importlib_resources.files(f"data.doc.cli").joinpath(
                        description_path
                    ),
                    "r",
                ).read()
            ),
            subtitle=f"[{citros_version}]",
            title="description",
        ),
        epilog=epilog(),
        formatter_class=RichHelpFormatterCitros,
        # add_help=False
    )

    parser.add_argument(
        "-dir", "--dir", default=".", help="The working dir of the project"
    )
    parser.add_argument(
        "-d", "--debug", action="store_true", help="set logging level to debug"
    )
    parser.add_argument(
        "-v", "--verbose", action="store_true", help="use verbose console prints"
    )
    parser.add_argument("-V", "--version", action="version", version=citros_version)
    parser.set_defaults(func=citros)

    def func(args, argv):
        print(PANNEL)
        citros(args, argv)

    parser.set_defaults(func=func)

    subparsers = parser.add_subparsers(
        title="commands",
        help="citros commands",
        dest="type",  # required=True
    )

    parser_init(
        subparsers,
        epilog("https://docs.citros.io/docs/next/commands/init"),
    )
    parser_run(
        subparsers,
        epilog("https://docs.citros.io/docs/next/commands/run"),
    )
    parser_data(
        subparsers,
        epilog("https://docs.citros.io/docs/next/commands"),
    )
    parser_report(
        subparsers,
        epilog(),
    )
    # parser_service(
    #     subparsers,
    #     epilog(),
    # )

    # parser_simulation(
    #     subparsers,
    #     epilog(),
    # )
    # parser_parameter(
    #     subparsers,
    #     epilog(),
    # )
    # parser_launch(
    #     subparsers,
    #     epilog(),
    # )

    args, argv = parser.parse_known_args()

    args.func(args, argv)


if __name__ == "__main__":
    main()



import argparse
import importlib_resources
from bin.cli_impl import *
from rich import print, inspect, print_json
from rich.rule import Rule
from rich.panel import Panel
from rich.padding import Padding
from rich.logging import RichHandler
from rich.console import Console
from rich.markdown import Markdown
from rich.traceback import install
from bin import __version__ as citros_version
from .formatter import RichHelpFormatterCitros

install()


# citros simulation run
def parser_simulation_run(subparser, epilog=None):
    description_path = "simulation/run.md"
    help = "citros data create section"
    parser = subparser.add_parser(
        "run",
        description=Panel(
            Markdown(
                open(
                    importlib_resources.files(f"data.doc.cli").joinpath(
                        description_path
                    ),
                    "r",
                ).read()
            ),
            subtitle=f"[{citros_version}]",
            title="description",
        ),
        epilog=epilog,
        help=help,
        formatter_class=RichHelpFormatterCitros,
    )
    parser.add_argument(
        "-d", "--debug", action="store_true", help="set logging level to debug"
    )
    parser.add_argument(
        "-v", "--verbose", action="store_true", help="use verbose console prints"
    )
    parser.set_defaults(func=simulation_run)

    return parser


# citros simulation list
def parser_simulation_list(subparser, epilog=None):
    description_path = "simulation/list.md"
    help = "citros data status section"
    parser = subparser.add_parser(
        "list",
        description=Panel(
            Markdown(
                open(
                    importlib_resources.files(f"data.doc.cli").joinpath(
                        description_path
                    ),
                    "r",
                ).read()
            ),
            subtitle=f"[{citros_version}]",
            title="description",
        ),
        epilog=epilog,
        help=help,
        formatter_class=RichHelpFormatterCitros,
    )
    parser.add_argument(
        "-d", "--debug", action="store_true", help="set logging level to debug"
    )
    parser.add_argument(
        "-v", "--verbose", action="store_true", help="use verbose console prints"
    )
    parser.set_defaults(func=simulation_list)
    return parser


# citros simulation
def parser_simulation(subparser, epilog=None):
    description_path = "simulation.md"
    help = "simulation section"
    parser = subparser.add_parser(
        "simulation",
        description=Panel(
            Markdown(
                open(
                    importlib_resources.files(f"data.doc.cli").joinpath(
                        description_path
                    ),
                    "r",
                ).read()
            ),
            subtitle=f"[{citros_version}]",
            title="description",
        ),
        epilog=epilog,
        help=help,
        formatter_class=RichHelpFormatterCitros,
        aliases=["sim"],
    )
    parser.add_argument("-n", "--name", default=None, help="name of simulation")
    parser.add_argument("-m", "--match", default=None, help="match simulation pattern")

    subparser = parser.add_subparsers(dest="type")
    # simulation run/list
    parser_simulation_run(subparser, epilog=epilog)
    parser_simulation_list(subparser, epilog=epilog)

    return parser



from rich_argparse import RichHelpFormatter


class RichHelpFormatterCitros(RichHelpFormatter):
    def __init__(
        self,
        prog: str,
        indent_increment: int = 2,
        max_help_position: int = 24,
        width: int | None = None,
    ) -> None:
        super().__init__(prog, indent_increment=4, max_help_position=40, width=width)



import os
import re
import csv
import git
import json
import time
import psutil
import shutil
import base64
import hashlib
import jsonschema
import importlib_resources

from os import linesep
from pathlib import Path
from datetime import datetime


################################## hash ###################################


def compute_sha256_hash(file_path):
    """
    Computes the SHA-256 hash of a file.
    """
    sha256_hash = hashlib.sha256()

    with open(file_path, "rb") as f:
        # Read and update hash in chunks of 4K
        for byte_block in iter(lambda: f.read(4096), b""):
            sha256_hash.update(byte_block)

    return sha256_hash.hexdigest()


################################## ssh ####################################


def check_ssh_key_pair():
    """
    Checks the existence of SSH key pairs in the user's SSH directory.

    Returns:
        str or None: The type of the existing key pair ('id_ed25519' or 'id_rsa') if found,
        or None if no key pair is found.
    """
    ssh_dir = Path("~/.ssh").expanduser()

    if (
        Path(ssh_dir, "citros_ed25519").exists()
        and Path(ssh_dir, "citros_ed25519.pub").exists()
    ):
        return "citros_ed25519"
    elif (
        Path(ssh_dir, "citros_rsa").exists()
        and Path(ssh_dir, "citros_rsa.pub").exists()
    ):
        return "citros_rsa"
    else:
        return None


############################## Network ####################################


def suppress_ros_lan_traffic(ros_domain_id):
    """
    avoid seeing ros traffic from other simulations on the same LAN.
    """
    if "ROS_DOMAIN_ID" not in os.environ:
        # anything between 0 and 101
        os.environ["ROS_DOMAIN_ID"] = str(ros_domain_id)


########################### file and format utils #########################


def is_valid_file_name(name: str, log):
    # check for empty name, invalid characters and trailing periods or spaces.
    if (
        not name
        or re.search(r'[\\/*?:,;"\'<>|(){}\t\r\n]', name)
        or name[-1] == "."
        or name[-1] == " "
    ):
        log.warning(f"invalid file or folder name: {name}")
        return False

    return True


def get_foramtted_datetime():
    now = datetime.now()
    formatted = now.strftime("%Y-%m-%d-%H-%M-%S")

    # Use only the last two digits of the year
    formatted = formatted[2:]

    return formatted


def str_to_bool(s):
    if s == "True":
        return True
    elif s == "False":
        return False
    else:
        raise ValueError(f"Cannot convert {s} to a bool")


# files
def get_data_file_path(data_package, filename):
    if data_package not in [
        "schemas",
        "defaults",
        "scripts",
        "sample_code",
        "markdown",
        "misc",
    ]:
        raise ValueError(f"data package '{data_package}' is unsupported.")

    return importlib_resources.files(f"data.{data_package}").joinpath(filename)


def copy_files(file_paths, target_directory, log, create_dir=False):
    if create_dir:
        # Create the target directory if it does not exist
        os.makedirs(target_directory, exist_ok=True)

    for file_path in file_paths:
        if os.path.isfile(file_path):
            shutil.copy(file_path, target_directory)
        else:
            log.error(f"copy_files: File does not exist: {file_path}")


def copy_subdir_files(source_directory, target_directory, log):
    """
    Copies files from the source to the target, if and only if the target has the
    same directory structure as the source, for any specific file.
    """
    if not os.path.exists(source_directory):
        log.error(f"Source directory does not exist: {source_directory}")
        return

    # Iterate through the subdirectories in the source directory
    for root, subdirs, files in os.walk(source_directory):
        relative_path = os.path.relpath(root, source_directory)
        target_subdir = os.path.join(target_directory, relative_path)

        # Check if the subdirectory exists in the target directory
        if os.path.exists(target_subdir):
            # List of file paths in the current subdirectory
            file_paths = [os.path.join(root, file) for file in files]
            copy_files(file_paths, target_subdir)
        else:
            log.error(f"Target subdirectory does not exist: {target_subdir}")


def get_last_created_file(folder_path, dirs=False):
    folder = Path(folder_path)
    items = []

    if dirs:
        items = [f for f in folder.iterdir() if f.is_dir()]
    else:
        items = [f for f in folder.iterdir() if f.is_file()]

    if not items:
        return None

    # Sort files by creation time
    items.sort(key=lambda x: x.stat().st_ctime)

    # The last item in the list is the most recently created file
    return items[-1]


def rename_file(file_path, new_name):
    file = Path(file_path)

    # Construct the new path
    new_file_path = file.parent / new_name

    # Rename the file
    file.rename(new_file_path)


def find_ancestor_with_name(path, name):
    """
    Checks all ancestors of the given path to see if one of them has the specified name.

    :param path: The path to start from.
    :param name: The name of the ancestor directory to find.
    :return: The Path of the ancestor directory if found, else None.
    """
    path = Path(path)
    for ancestor in path.parents:
        if ancestor.name == name:
            return ancestor
    return None


########################### git #########################


def update_git_exclude(repo_path, pattern, log):
    gitexclude_path = Path(repo_path, ".git/info/exclude")
    if not gitexclude_path.exists():
        log.warning(f"Could not find git exclude in repo {repo_path}")
        return

    def normalize_pattern(pattern: str):
        # Remove any trailing slash or asterisk
        return pattern.rstrip("*/")

    normalized_pattern = normalize_pattern(pattern)

    with open(gitexclude_path, "r+") as gitexclude:
        lines = gitexclude.readlines()
        for line in lines:
            if normalize_pattern(line.strip()) == normalized_pattern:
                # pattern already exists.
                return

        # Pattern not found, append it
        gitexclude.write(linesep + pattern + linesep)
        print(f"Pattern `{pattern}` appended to {gitexclude_path}")


def get_git_info(self, repo_path="."):
    try:
        repo = git.Repo(repo_path)
    except:
        return None, None

    try:
        latest_commit_hash = repo.head.commit.hexsha
    except:
        # empty repo (https://github.com/lulav/citros_cli/issues/51)
        latest_commit_hash = None

    current_branch_name = repo.active_branch.name

    return latest_commit_hash, current_branch_name


def get_user_git_info(dir):
    commit, branch = get_git_info(dir)
    return commit, branch


########################### Validation #########################


def _has_files(dir: Path):
    return any(p.is_file() for p in dir.iterdir())


def _validate_json_file(json_filepath, schema_filepath, log):
    with open(json_filepath, "r") as file:
        data = json.load(file)

    with open(schema_filepath, "r") as file:
        schema = json.load(file)

    try:
        # in order to support future changes in schemas (i.e. additional fields),
        # we merge with the default, so any new fields will be added to the user's
        # file and the validation will succeed.
        merged_dict = data  # {**default, **data}
        jsonschema.validate(instance=merged_dict, schema=schema)
        # print(f"{json_filepath} is valid.")

        if merged_dict != data:
            with open(json_filepath, "w") as file:
                file.write(json.dumps(merged_dict, indent=4))

        return True
    except jsonschema.exceptions.ValidationError as ve:
        log.error(f"{json_filepath} is not valid:{linesep}{ve}")
        return False


def validate_file(file_path: Path, schema_file, log):
    if file_path.is_file():
        with get_data_file_path("schemas", schema_file) as schema_path:
            return _validate_json_file(str(file_path), schema_path, log)
    else:
        return False


def validate_dir(path: Path, schema_file, log):
    if path.is_dir() and _has_files(path):
        with get_data_file_path("schemas", schema_file) as schema_path:
            for file in path.iterdir():
                if str(file).endswith("json"):
                    if not _validate_json_file(file, schema_path, log):
                        return False
            return True
    else:
        # empty folder is valid
        return True



import os
import time
import json
import yaml
import logging
import psycopg2
import traceback

from glob import glob
from rich import print, inspect, print_json
from jinja2 import Environment, FileSystemLoader
from pathlib import Path
from decouple import config


class NoConnectionToCITROSDBException(Exception):
    def __init__(self, message="No connection to citros db."):
        super().__init__(message)


class BatchUploader:
    """
    The Uploader class is responsible for uploading data to CITROS db.


    assuming structure:
        .citros/data/{simulation name}/{batch name}/{simulation run id}/bag/{bag file}

    Attributes:
    -----------
    root : str
        The root directory where the data is located. (default to .citros/data)
    simulation : str
        The simulation name.
    batch : str
        The batch name.
    version : str
        The version of the uploader.

    log : Optional
        The log object for logging messages.
    verbose : bool
        Flag indicating whether to enable verbose mode.
    debug : bool
        Flag indicating whether to enable debug mode.

    """

    ###################
    ##### private #####
    ###################

    def _get_citros_bags(self, path: str) -> list:
        """
        Search for `.citros.mcap` and `.citros.db3` bag files in a specified directory.

        This method walks through the directory structure rooted at `path` to locate
        files ending with `.citros.mcap` or `.citros.db3`. It returns a list containing
        the full paths to these files.

        Parameters:
        -----------
        path : str
            The root directory where the search will begin.

        Returns:
        --------
        list
            A list of full file paths to the bag files. An exception is raised if no
            bag files are found.

        Raises:
        -------
        Exception
            If no `.citros.mcap` or `.citros.db3` bag files are found.

        Examples:
        ---------
        >>> _get_citros_bags('/path/to/root')
        ['/path/to/root/folder1/bag1.citros.mcap', '/path/to/root/folder2/bag2.citros.db3']

        """
        bags = []
        for root, dirs, files in os.walk(path):
            for file in files:
                if file.endswith(".citros.mcap"):
                    self.log.debug(f"Found MCAP bag: {file}")
                elif file.endswith(".citros.db3"):
                    self.log.debug(f"Found SQLITE3 bag: {file}")
                else:
                    continue
                bags.append(os.path.join(root, file))
        # if len(bags) < 1:
        #     raise Exception(
        #         f"Didn't find SQLITE3 or MCAP bag in the [{path}] folder ..."
        #     )

        return bags

    def _get_custom_message(self, directory_path: str) -> list:
        """
        Search for and return the paths of folders containing custom message types.

        This method searches through the directory specified by `directory_path` to
        find any folders that contain custom message types. These folders are assumed
        to contain a 'msg/' subdirectory.

        Parameters:
        -----------
        directory_path : str
            The directory path where the method will search for custom message types.

        Returns:
        --------
        list
            A list of strings, each being the path of a folder containing custom message types.

        Raises:
        -------
        Exception
            If any error occurs while accessing the file system.

        Examples:
        ---------
        >>> _get_custom_message('/path/to/directory')
        # ['/path/to/directory/custom_msgs1/msg/', '/path/to/directory/custom_msgs2/msg/']

        """
        try:
            packages = [
                folder
                for folder in os.listdir(directory_path)
                if os.path.isdir(os.path.join(directory_path, folder))
            ]
            msgs = [directory_path + package + "/msg/" for package in packages]
            if len(msgs) > 0:
                self.log.debug(f"Found custom message in {msgs}")
        except Exception:
            msgs = []
        return msgs

    def _get_parameters(self, directory_path: str) -> list:
        """
        Search for and return the paths of parameter files in the given directory.

        This method scans the directory specified by `directory_path` to find parameter files.
        It looks for files ending with ".json", ".yaml", or ".yml".

        Parameters:
        -----------
        directory_path : str
            The directory path where the method will search for parameter files.

        Returns:
        --------
        list
            A list of strings, each being the path of a parameter file.

        Logs:
        -----
        Debug log if no parameters are found in the specified directory.

        Examples:
        ---------
        >>> _get_parameters('/path/to/directory')
        # ['/path/to/directory/params1.json', '/path/to/directory/params2.yaml']

        """
        parameters = []
        for root, dirs, files in os.walk(directory_path):
            for file in files:
                if file.endswith(".json"):
                    self.log.debug(f"Found json parameters file: {file}")
                elif file.endswith(".yaml") or file.endswith(".yml"):
                    self.log.debug(f"Found yaml parameters file: {file}")
                else:
                    continue
                parameters.append(os.path.join(root, file))
        if len(parameters) < 1:
            self.log.debug(
                f"Didn't find parameters in the [{directory_path}] folder ..."
            )

        return parameters

    @staticmethod
    def _read_parameters_file(file_path: str) -> dict:
        """
        Read the content of a parameter file and return it as a JSON-formatted dictionary.

        This method reads a parameter file specified by `file_path`, converts its content to
        a Python dictionary, and then returns the dictionary as a JSON-formatted string.
        It can read both JSON and YAML formatted parameter files.

        Parameters:
        -----------
        file_path : str
            The path to the parameter file to be read. The file should have either a .json, .yaml, or .yml extension.

        Returns:
        --------
        dict
            A dictionary containing the parameters read from the file, serialized to JSON format.

        Raises:
        -------
        Exception if the file cannot be read or parsed.

        Examples:
        ---------
        >>> _read_parameters_file('/path/to/params.json')
        # '{"key": "value"}'

        >>> _read_parameters_file('/path/to/params.yaml')
        # '{"key": "value"}'

        """
        _, ext = os.path.splitext(file_path)
        with open(file_path, "r") as f:
            if ext == ".json":
                return json.dumps(json.load(f))
            elif ext == ".yaml" or ext == ".yml":
                return json.dumps(yaml.safe_load(f))

    def upload_bag_to_pg(
        self,
        connection,
        schema_name: str,
        table_name: str,
        sid: int,
        bag: str,
        msgs: list,
    ):
        """
        Upload a bag file to a PostgreSQL database.

        This method reads messages from a bag file using the `BagReaderCustomMessages` class and uploads them to
        a PostgreSQL database table. The method handles both successful and unsuccessful uploads, logging
        appropriate messages in both cases.

        Parameters:
        ----------
        cursor : psycopg2.cursor
            The cursor object for interacting with the PostgreSQL database.

        connection : psycopg2.connection
            The connection object for the PostgreSQL database.

        schema_name : str

        table_name : str

        sid : str
            The ID of the simulation run to which this bag file belongs.

        bag : str
            The path to the bag file that needs to be uploaded.

        msgs : list
            List of paths to custom message directories (if any).

        Returns:
        -------
        tuple
            A tuple containing:
            - A boolean indicating whether the upload was successful.
            - A status message string.
            - None if successful, or an error message string if unsuccessful.

        Raises:
        ------
        Exception, psycopg2.Error:
            Any exceptions raised during database interactions are caught and logged.

        Examples:
        --------
        >>> upload_bag_to_pg(cursor, connection, 'batch_001', 'sim_001', '/path/to/bag', ['/path/to/custom/msg'])
        # (True, 'Success, uploaded [/path/to/bag] to Postgres. [size: 100 MB]', None)
        """
        from .ros import BagReaderCustomMessages

        cursor = connection.cursor()
        try:
            start_time = time.time()
            self.log.info(f"Uploading {bag} to PG")
            bagReader = BagReaderCustomMessages(
                msgs,
                log=self.log,
                debug=self.debug,
                verbose=self.verbose,
            )
            total_size = 0
            for buffer in bagReader.read_messages(bag, sid):
                size = buffer.seek(0, os.SEEK_END)
                size = buffer.tell()
                total_size = total_size + size
                buffer.seek(0)
                self.log.debug(
                    f" \tInserting buffer size: {((size / 1024 ) / 1024):.2f} MB"
                )
                if size == 0:
                    continue
                cursor.execute(f"SET search_path TO {schema_name}")
                try:
                    cursor.copy_from(
                        buffer,
                        table_name,
                        sep=chr(0x1E),
                        null="",
                        columns=["sid", "rid", "time", "topic", "type", "data"],
                    )
                except (Exception, psycopg2.Error) as error:
                    buffer.seek(0)
                    self.log.error(f"buffer = {buffer.getvalue()}")
                    self.log.error(
                        f" Failed to insert record into table, aborting upload to DB.",
                        error,
                    )
                    self.log.exception(error)
                    return False, "Got exception from pgdb", str(error)
                connection.commit()

            self.log.debug(
                f"Done uploading {bag}, took {(time.time() - start_time):.3f} [sec]"
            )
            return (
                True,
                f"Success, uploaded [{bag}] to Postgres. [size: {(total_size / 1024)/1024} MB]",
                None,
            )
        except (Exception, psycopg2.Error) as error:
            connection.commit()
            self.log.exception(
                f" Failed to insert record into table, aborting upload to DB.", error
            )
            self.log.exception(traceback.format_exc())
            return False, "Got exception from pgdb", str(error)
        # finally:
        #     # closing database connection.
        #     if connection:
        #         cursor.close()
        #         connection.close()
        #         self.log.debug(f"PostgreSQL connection is closed")
        #         logging.shutdown()

    def upload_parameters_to_pg(
        self,
        connection,
        schema_name: str,
        table_name: str,
        sid: str,
        parameter_file: str,
    ):
        """
        Upload simulation parameters to a PostgreSQL database table.

        This function takes a dictionary of parameters and uploads them to a specified table in a PostgreSQL database.
        If the upload is successful, the method will commit the changes and return a success message.

        Parameters:
        ----------

        connection : psycopg2.connection
            The connection object for the PostgreSQL database.

        params_dict : dict
            A dictionary containing key-value pairs of parameters to upload. Keys are parameter names as strings,
            and values are the corresponding parameter values.

        table_name : str, optional
            The name of the database table where parameters should be uploaded. Default is 'parameters'.

        Returns:
        -------
        tuple
            A tuple containing:
            - A boolean indicating whether the upload was successful.
            - A status message string.
            - None if successful, or an error message string if unsuccessful.

        Raises:
        ------
        psycopg2.Error:
            Any exceptions raised during database interactions are caught and logged.

        Examples:
        --------
        >>> params = {'param1': 1, 'param2': 2}
        >>> upload_parameters_to_pg(cursor, connection, params)
        # (True, 'Parameters uploaded successfully to table [parameters].', None)
        """
        try:
            self.log.debug(f"Uploading parameter file: {parameter_file}")
            parameter_dict = self._read_parameters_file(parameter_file)
            self.log.debug(f"Parameter:\n[{json.dumps(parameter_dict, indent=2)}]")
            record_to_insert = {
                "sid": sid,
                "rid": 0,
                "time": 0,
                "topic": "/config",
                "type": parameter_file,
                "data": parameter_dict,
            }

            postgres_insert_query = f""" 
                insert into {schema_name}."{table_name}"
                (sid, rid, time, topic, type, data)
                values (%(sid)s, %(rid)s, %(time)s, %(topic)s, %(type)s, %(data)s);
            """

            cursor = connection.cursor()
            cursor.execute(postgres_insert_query, record_to_insert)
            connection.commit()

        except Exception as e:
            self.log.error(e)

        # finally:
        #     if connection:
        #         cursor.close()
        #         connection.close()
        #         self.log.debug(f"PostgreSQL connection is closed")
        #         logging.shutdown()

    def upload(self):
        self.log.debug(f"{self.__class__.__name__}.upload()")

        # inspect(self)
        # print(f"self.batch_dir: {self.batch_dir}")

        from .database import CitrosDB

        citrosDB = CitrosDB(log=self.log, debug=self.debug, verbose=self.verbose)

        schema_name = f"{self.simulation_name}"
        table_name = f"{self.name}"

        connection = citrosDB.connect()
        if connection is None:
            self.log.error("No creating connection to database. Aborting.")
            raise NoConnectionToCITROSDBException

        citrosDB.create_table(connection, schema_name, table_name, self.version)

        for sid_path in glob(f"{self.batch_dir}/*/"):
            if sid_path.endswith("/"):
                sid_path = sid_path[:-1]

            sid = sid_path.split("/")[-1]

            # upload all parameters
            parameters = self._get_parameters(f"{self.batch_dir}/{sid}/config/")
            # print(parameters)
            for parameter in parameters:
                try:
                    # cursor = connection.cursor()
                    self.upload_parameters_to_pg(
                        connection,
                        schema_name=schema_name,
                        table_name=table_name,
                        sid=sid,
                        parameter_file=parameter,
                    )
                    connection.commit()
                except Exception as e:
                    self.log.error(e)
                    connection = citrosDB.connect()

            # uplaod bags
            bags = self._get_citros_bags(f"{self.batch_dir}/{sid}/bags/")
            if len(bags) < 1:
                self.log.error(
                    f"No bags found in {self.batch_dir}/{sid}/bags/ . Perhaps the run didnt finish properly."
                )
                continue

            # print(bags)
            msgs = self._get_custom_message(f"{self.batch_dir}/{sid}/msgs/")
            # print(msgs)

            for bag in bags:
                try:
                    status, message, error = self.upload_bag_to_pg(
                        connection,
                        schema_name=schema_name,
                        table_name=table_name,
                        sid=sid,
                        bag=bag,
                        msgs=msgs,
                    )
                    self.log.debug(
                        f"status: {status}, message: {message}, error: {error}"
                    )
                except Exception as e:
                    self.log.error(e)
                    connection = citrosDB.connect()

        citrosDB.hot_reload_set_status(
            connection, schema_name, table_name, self.version, "LOADED"
        )
        # cursor.close()
        connection.commit()
        connection.close()

    def unload(self):
        from .database import CitrosDB

        citrosDB = CitrosDB(log=self.log, debug=self.debug, verbose=self.verbose)

        schema_name = f"{self.simulation_name}"
        table_name = f"{self.name}"

        connection = citrosDB.connect()
        if connection is None:
            self.log.error("No creating connection to database. Aborting.")
            raise NoConnectionToCITROSDBException

        citrosDB.drop_table(connection, schema_name, table_name)

        connection.close()



import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import numpy as np

from citros import CitrosDB, CitrosData, CitrosDataArray
from decouple import config
'''
tests for error_analysis module

just type:

pytest

in terminal in this folder
'''
#connect to a database
table_name = 'test_table'

if config('TEST_ENV', None) == 'github':
    citros = CitrosDB(simulation = 'public',
                      batch = table_name,
                      host=config('POSTGRES_HOST'),
                      user=config('POSTGRES_USER'),
                      password=config('POSTGRES_PASSWORD'),
                      database=config('POSTGRES_DB'),
                      port = config('POSTGRES_PORT'))
else:
    citros = CitrosDB(simulation = 'public',
                        batch = table_name)

citros_D = CitrosData()

x_1 = [5.0, -2.0, -10.0, -2.0, 8.0, -10.0, 0.0, 4.0, -9.0, 5.0, 4.0, np.nan, 8.0, 1.0, -1.0]
x_2 = [0.91, 0.27, 0.27, 0.58, 0.7, 0.21, 0.77, 0.26, 0.74, 0.79, 0.42, 0.03, 0.27, 0.98, 0.71]
x_3 = [-1.4, 1.6, -0.5, 0.9, 9.7, 2.6, -0.8, -6.0, -6.0, 3.3, 4.7, -8.9, 5.9, np.nan, -5.1]
F=pd.DataFrame({'x_1': x_1, 'x_2': x_2, 'x_3': x_3})

X = np.array([[4,17,4],[11,10,12],[6,9,2]])

def test_get_mu():
    mu = citros_D._get_mu(X)
    mu_expected = np.array([7, 12, 6])
    for val in abs(mu -mu_expected):
        assert val < 0.00001, "mu from array is wrong"
    
    mu_2 = citros_D._get_mu(F[['x_1','x_2','x_3']].iloc[9:])
    mu_2_expected = np.array([4.0, 0.5475, 2.2])
    for val in abs(mu_2 -mu_2_expected):
        assert val < 0.00001, "mu from data frame is wrong"

def test_get_covar_matrix():
    sigm = citros_D._get_covar_matrix(X)
    sigm_expected = np.array([[13,-10,17],[-10,19,-5],[17,-5,28]])
    for i in range(3):
        for j in range(3):
            assert abs(sigm[i][j]-sigm_expected[i][j]) < 0.00001, "covariance matrix from array is wrong"

    sigm_2 = citros_D._get_covar_matrix(F[['x_1','x_2','x_3']].iloc[9:])
    sigm_2_expected = np.array([[14.0000, -0.5600, 17.46667],[-0.5600, 0.0594917, -0.7550,],[17.46667,-0.7550,24.81333]])
    for i in range(3):
        for j in range(3):
            assert abs(sigm_2[i][j]-sigm_2_expected[i][j]) < 0.00001, "covariance matrix from data frame is wrong"

def test_get_disp():
    disp = citros_D._get_disp(X)
    disp_expected = [3.60555, 4.358899, 5.291503] 
    for val in abs(disp - disp_expected):
        assert val < 0.00001, "dispersion from array is wrong"

    disp_2 = citros_D._get_disp(F[['x_1','x_2','x_3']].iloc[9:])
    disp_2_expected =[3.741657, 0.243909, 4.981298]
    for val in abs(disp_2 - disp_2_expected):
        assert val < 0.00001, "dispersion from data frame is wrong"

def test_CitrosData():
    #construct CitrosDataSet object with 3 data-columns from 'data.x', parameter from the column 'data.p'
    df = citros.topic('A').data(['data.x', 'data.time', 'data.p'])
    dataset = CitrosData(df, type_name = 'x', data_label=['data.x'], units = 'm', parameter_label='data.p')
    assert len(dataset.data.columns) == 3, 'CitrosData: wrong number of columns'
    assert dataset.units == 'm', 'CitrosData: wrong units'
    assert list(dataset.parameters.keys())[0] == 'data.p', 'CitrosData: wrong parameter_label'
    assert dataset.type == 'x', 'CitrosData: wrong type'

    #construct CitrosDataSet object with 2 data-columns, parameter from the column 'data.p'
    df = citros.topic('A').data(['data.x.x_1','data.x.x_2', 'data.time', 'data.p', 'data.t'])
    dataset = CitrosData(df, type_name = 'x', data_label=['data.x.x_1','data.x.x_2'], units = 'm')
    dataset.set_parameter('p', 5.5)
    dataset.set_parameter('k', 8)
    assert len(dataset.data.columns) == 2, 'CitrosData: wrong number of columns'
    assert dataset.parameters['p'] == 5.5 and dataset.parameters['k'] == 8, 'CitrosData: parameters set wrong'
    
    #construct CitrosDataSet object with 1 data-columns, parameter from 2 columns 'data.p' and 'data.t'
    dataset = CitrosData(df, type_name = 'x', data_label=['data.x.x_1'], units = 'm', parameter_label=['data.p','data.t'])
    assert len(dataset.data.columns) == 1, 'CitrosData: wrong number of columns'
    assert ('data.p' in list(dataset.parameters.keys())) and ('data.t' in list(dataset.parameters.keys())), 'CitrosData: wrong parameter_label'
    
    #test filter on columns with nan and inf == +-10**308 values
    df = citros.topic('A').data(['data.y.y_1', 'data.y.y_2'])
    dataset = CitrosData(df, data_label = ['data.y.y_1','data.y.y_2'], units = 'm')
    assert (((dataset.data['data.y.y_1'].abs()-10**307)>=0) | dataset.data['data.y.y_1'].isna() | 
            ((dataset.data['data.y.y_2'].abs()-10**307)>=0) | dataset.data['data.y.y_2'].isna())\
                .value_counts()[True] == dataset.filter['data.y.y_2'].value_counts()[False], \
                'CitrosData: omit_nan_rows, inf_vals: wrong filter'
    assert (dataset.filter['data.y.y_2'] == dataset.filter['data.y.y_1']).all() == True, \
        'CitrosData: omit_nan_rows, inf_vals: filters are not equal'

    dataset = CitrosData(df, data_label = ['data.y.y_1','data.y.y_2'], units = 'm', inf_vals = None)
    assert (dataset.data['data.y.y_1'].isna() | dataset.data['data.y.y_2'].isna()).\
        value_counts()[True] == dataset.filter['data.y.y_2'].value_counts()[False], \
        'CitrosData: omit_nan_rows: wrong filter'
    assert (dataset.filter['data.y.y_2'] == dataset.filter['data.y.y_1']).all() == True, \
        'CitrosData: omit_nan_rows: filters are not equal'

    dataset = CitrosData(df, data_label = ['data.y.y_1','data.y.y_2'], units = 'm', omit_nan_rows=False)
    assert (((dataset.data['data.y.y_1'].abs()-10**307)>=0) | dataset.data['data.y.y_1'].isna()).\
        value_counts()[True] == dataset.filter['data.y.y_1'].value_counts()[False], \
        'CitrosData: inf_vals: wrong filter for the first column'
    assert (((dataset.data['data.y.y_2'].abs()-10**307)>=0) | dataset.data['data.y.y_2'].isna()).\
        value_counts()[True] == dataset.filter['data.y.y_2'].value_counts()[False], \
        'CitrosData: inf_vals: wrong filter for the second column'

    dataset = CitrosData(df, data_label = ['data.y.y_1','data.y.y_2'], units = 'm', inf_vals=None, omit_nan_rows=False)
    assert (dataset.data['data.y.y_1'].isna()).value_counts()[True] == dataset.filter['data.y.y_1'].value_counts()[False], \
        'CitrosData: no inf_vals, omit_nan_rows=False : wrong filter for the first column'
    assert (dataset.data['data.y.y_2'].isna()).value_counts()[True] == dataset.filter['data.y.y_2'].value_counts()[False], \
        'CitrosData: no inf_vals, omit_nan_rows=False : wrong filter for the second column'

def test_bin_data():
    #test_bin_data
    df = citros.topic('A').rid(start = 0, end = 10).data(['data.x', 'data.time', 'data.p'])
    dataset = CitrosData(df, type_name = 'x', data_label=['data.x'], units = 'm')
    db_bin=dataset.bin_data(n_bins = 5, param_label = 'data.time')

    time_min = min(dataset.addData['data.time'])
    time_max = max(dataset.addData['data.time'])
    time_bin = np.linspace(time_min, time_max, 6)
    time_bin_centers_answ = [(time_bin[i-1]+time_bin[i])/2 for i in range(1,len(time_bin))]

    time_bin_centers = list(set(db_bin.addData['data.time']))
    time_bin_centers.sort()

    assert len(time_bin_centers)==len(time_bin_centers_answ), 'wrong bin centers length'
    assert all([abs(t-answ)<0.01 for t, answ in zip(time_bin_centers, time_bin_centers_answ)]), 'wrong bin centers'

    flag_bin1 = (dataset.addData['data.time'] >= time_bin[0]) & (dataset.addData['data.time'] <= time_bin[1])
    assert len(dataset.addData[flag_bin1]) == 10, 'number of counts in the first bin is wrong'

    assert abs(db_bin.data.loc[0].loc[1].subtract(dataset.data[flag_bin1 & (dataset.addData['sid']==1)].mean()).sum()) < 0.001, 'data values in the first bin, sid = 1 is wrong'
    assert abs(db_bin.data.loc[0].loc[2].subtract(dataset.data[flag_bin1 & (dataset.addData['sid']==2)].mean()).sum()) < 0.001, 'data values in the first bin, sid = 2 is wrong'
    assert abs(db_bin.data.loc[0].loc[3].subtract(dataset.data[flag_bin1 & (dataset.addData['sid']==3)].mean()).sum()) < 0.001, 'data values in the first bin, sid = 3 is wrong'

    flag_bin2 = (dataset.addData['data.time'] > time_bin[1]) & (dataset.addData['data.time'] <= time_bin[2])
    assert len(dataset.addData[flag_bin2]) == 11, 'number of counts in the second bin is wrong'
    assert abs(db_bin.data.loc[1].loc[1].subtract(dataset.data[flag_bin2 & (dataset.addData['sid']==1)].mean()).sum()) < 0.001, 'data values in the second bin, sid = 1 is wrong'
    assert abs(db_bin.data.loc[1].loc[2].subtract(dataset.data[flag_bin2 & (dataset.addData['sid']==2)].mean()).sum()) < 0.001, 'data values in the second bin, sid = 2 is wrong'
    assert abs(db_bin.data.loc[1].loc[3].subtract(dataset.data[flag_bin2 & (dataset.addData['sid']==3)].mean()).sum()) < 0.001, 'data values in the second bin, sid = 3 is wrong'

    flag_bin3 = (dataset.addData['data.time'] > time_bin[2]) & (dataset.addData['data.time'] <= time_bin[3])
    assert len(dataset.addData[flag_bin3]) == 7, 'number of counts in the third bin is wrong'
    assert abs(db_bin.data.loc[2].loc[1].subtract(dataset.data[flag_bin3 & (dataset.addData['sid']==1)].mean()).sum()) < 0.001, 'data values in the third bin, sid = 1 is wrong'
    assert abs(db_bin.data.loc[2].loc[2].subtract(dataset.data[flag_bin3 & (dataset.addData['sid']==2)].mean()).sum()) < 0.001, 'data values in the third bin, sid = 2 is wrong'
    assert abs(db_bin.data.loc[2].loc[3].subtract(dataset.data[flag_bin3 & (dataset.addData['sid']==3)].mean()).sum()) < 0.001, 'data values in the third bin, sid = 3 is wrong'

    flag_bin4 = (dataset.addData['data.time'] > time_bin[3]) & (dataset.addData['data.time'] <= time_bin[4])
    assert len(dataset.addData[flag_bin4]) == 2, 'number of counts in the fourth bin is wrong'
    assert abs(db_bin.data.loc[3].loc[1].subtract(dataset.data[flag_bin4 & (dataset.addData['sid']==1)].mean()).sum()) < 0.001, 'data values in the fourth bin, sid = 1 is wrong'
    assert abs(db_bin.data.loc[3].loc[2].subtract(dataset.data[flag_bin4 & (dataset.addData['sid']==2)].mean()).sum()) < 0.001, 'data values in the fourth bin, sid = 2 is wrong'

    flag_bin5 = (dataset.addData['data.time'] > time_bin[4]) & (dataset.addData['data.time'] <= time_bin[5])
    assert len(dataset.addData[flag_bin5]) == 3, 'number of counts in the fifth bin is wrong'
    assert abs(db_bin.data.loc[4].loc[1].subtract(dataset.data[flag_bin5 & (dataset.addData['sid']==1)].mean()).sum()) < 0.001, 'data values in the fifth bin, sid = 1 is wrong'

def test_bin_data_with_nan_1col():
    #tests bin with nan and inf vals, 1 column
    df = citros.topic('A').data(['data.y.y_1', 'data.y.y_2'])
    dataset = CitrosData(df, data_label = 'data.y.y_1', units = 'm')
    db = dataset.bin_data(n_bins = 1, param_label = 'data.y.y_2', show_fig = True)
    db.data.loc[0]

    flag_y = (((dataset.data['data.y.y_1'].abs()-10**308)<0) & dataset.data['data.y.y_1'].notna())
    flag_x = (((dataset.addData['data.y.y_2'].abs()-10**307)<0) & dataset.addData['data.y.y_2'].notna())
    flag = flag_x & flag_y
    assert (db.data.loc[0] == dataset.data[flag].groupby(dataset.addData[flag.values]['sid']).mean())['data.y.y_1'].all(),\
        'tests bin with nan and inf vals, 1 column: wrong result'
    assert (dataset.addData['data.y.y_2'][flag_x].max() + dataset.addData['data.y.y_2'][flag_x].min())/2 == db.addData['data.y.y_2'].loc[(0,1)], \
        'tests bin with nan and inf vals, 1 column: wrong result for x'

def test_bin_data_with_nan_2col_omit_nan_rows():
    #tests bin with nan and inf vals, omit_nan_rows=True, 2 columns
    df = citros.topic('A').data(['data.y.y_1', 'data.y.y_2', 'data.y.y_3'])
    dataset = CitrosData(df, data_label = ['data.y.y_1', 'data.y.y_3'], units = 'm', omit_nan_rows=True)
    db = dataset.bin_data(n_bins = 1, param_label = 'data.y.y_2', show_fig = False)

    flag_y = (((dataset.data['data.y.y_1'].abs()-10**308)<0) & dataset.data['data.y.y_1'].notna())
    flag_y = flag_y & (((dataset.data['data.y.y_3'].abs()-10**308)<0) & dataset.data['data.y.y_3'].notna())
    flag_x = (((dataset.addData['data.y.y_2'].abs()-10**308)<0) & dataset.addData['data.y.y_2'].notna())
    flag = flag_x & flag_y
    assert (db.data.xs(0, level = db.xid_label) == dataset.data[flag].groupby(dataset.addData[flag.values]['sid']).mean())['data.y.y_1'].all(), \
        'tests bin with nan and inf vals, 2 columns, omit_nan_rows=True: wrong result for the first column'
    assert (db.data.xs(0, level = db.xid_label) == dataset.data[flag].groupby(dataset.addData[flag.values]['sid']).mean())['data.y.y_3'].all(), \
        'tests bin with nan and inf vals, 2 columns, omit_nan_rows=True: wrong result for the second column'
    assert (dataset.addData['data.y.y_2'][flag_x].max() + dataset.addData['data.y.y_2'][flag_x].min())/2 == db.addData['data.y.y_2'].loc[(0,1)], \
        'tests bin with nan and inf vals, 2 columns, omit_nan_rows=True: wrong result for x'

def test_bin_data_with_nan_1co2():
    #tests bin with nan and inf vals, omit_nan_rows=False, 2 columns
    df = citros.topic('A').data(['data.y.y_1', 'data.y.y_2', 'data.y.y_3'])
    dataset = CitrosData(df, data_label = ['data.y.y_1', 'data.y.y_3'], units = 'm', omit_nan_rows=False)
    db = dataset.bin_data(n_bins = 1, param_label = 'data.y.y_2', show_fig = True)

    flag_1y = (((dataset.data['data.y.y_1'].abs()-10**308)<0) & dataset.data['data.y.y_1'].notna())
    flag_3y = (((dataset.data['data.y.y_3'].abs()-10**308)<0) & dataset.data['data.y.y_3'].notna())
    flag_x = (((dataset.addData['data.y.y_2'].abs()-10**308)<0) & dataset.addData['data.y.y_2'].notna())
    flag_1 = flag_x & flag_1y
    flag_3 = flag_x & flag_3y
    assert (db.data['data.y.y_1'].loc[0] == dataset.data['data.y.y_1'][flag_1].groupby(dataset.addData[flag_1.values]['sid']).mean()).all(), \
        'tests bin with nan and inf vals, 2 columns, omit_nan_rows=False: wrong result for the first column'
    assert (db.data['data.y.y_3'].loc[0] == dataset.data['data.y.y_3'][flag_3].groupby(dataset.addData[flag_3.values]['sid']).mean()).all(), \
        'tests bin with nan and inf vals, 2 columns, omit_nan_rows=False: wrong result for the second column'
    assert (dataset.addData['data.y.y_2'][flag_x].max() + dataset.addData['data.y.y_2'][flag_x].min())/2 == db.addData['data.y.y_2'].loc[(0,1)], \
        'tests bin with nan and inf vals, 2 columns, omit_nan_rows=False: wrong result for x'

def test_scale_data():
    df = citros.topic('A').rid(start = 0, end = 10).data(['data.x', 'data.time', 'data.p'])
    dataset = CitrosData(df, type_name = 'x', data_label=['data.x'], units = 'm')
    db_sc = dataset.scale_data(n_points = 5, 
                                    param_label = 'data.time', 
                                    show_fig = False)

    time_points_answ = np.linspace(0, 1, 5)
    time_points = list(set(db_sc.addData['data.time']))
    time_points.sort()

    assert len(time_points)==len(time_points_answ), 'wrong number of points'
    assert all([abs(t-answ)<0.01 for t, answ in zip(time_points, time_points_answ)]), 'wrong points centers'

    def check_interpolation_value(n_time_point, n_sid, label):
        flag = dataset.addData['sid']==n_sid
        time_scaled = (dataset.addData['data.time'] -min(dataset.addData[flag]['data.time']))/(max(dataset.addData[flag]['data.time'])-min(dataset.addData[flag]['data.time']))
        index = abs(time_scaled[dataset.addData['sid']==n_sid] - time_points[n_time_point]).idxmin()

        assert abs(dataset.data.loc[index][label] - db_sc.data.loc[n_time_point].loc[n_sid][label])< 0.1*abs(dataset.data.loc[index][label]), \
            'error time_point = '+str(n_time_point)+' ,sid = '+str(n_sid)+', label = '+label

    check_interpolation_value(n_time_point = 1, n_sid = 2, label = 'data.x.x_1')
    check_interpolation_value(n_time_point = 3, n_sid = 3, label = 'data.x.x_1')
    check_interpolation_value(n_time_point = 3, n_sid = 1, label = 'data.x.x_1')
    check_interpolation_value(n_time_point = 1, n_sid = 3, label = 'data.x.x_2')
    check_interpolation_value(n_time_point = 4, n_sid = 2, label = 'data.x.x_2')

# @image_comparison(baseline_images=['scale_data_omit_nan_rows'], remove_text=True,
#                   extensions=['png'], style='mpl20')
# def test_scale_data_omit_nan_rows():
#     # with nan values and inf values = +-1e308
#     plt.close('all')
#     df = citros.topic('A').data(['data.y.y_1', 'data.y.y_2', 'data.y.y_3'])
#     dataset = CitrosData(df, data_label = ['data.y.y_1','data.y.y_3'], units = 'm', omit_nan_rows=True)
#     db = dataset.scale_data(n_points = 5, 
#                             param_label = 'data.y.y_2', 
#                             show_fig = True)
#     plt.close()
#     time_points_answ = np.linspace(0, 1, 5)
#     assert len(db.addData.groupby('data.y.y_2_id').first()['data.y.y_2'])==len(time_points_answ), 'wrong number of points'
#     assert all([abs(t-answ)<0.01 for t, answ in zip(db.addData.groupby('data.y.y_2_id').first()['data.y.y_2'], time_points_answ)]), 'wrong points centers'

# @image_comparison(baseline_images=['scale_data_with_nan_rows'], remove_text=True,
#                   extensions=['png'], style='mpl20')
# def test_scale_data_with_nan_rows():
#     # with nan values and inf values = +-1e308
#     plt.close('all')
#     df = citros.topic('A').data(['data.y.y_1', 'data.y.y_2', 'data.y.y_3'])
#     dataset = CitrosData(df, data_label = ['data.y.y_1','data.y.y_3'], units = 'm', omit_nan_rows=False)
#     db = dataset.scale_data(n_points = 5, 
#                             param_label = 'data.y.y_2', 
#                             show_fig = True)
#     plt.close()
#     time_points_answ = np.linspace(0, 1, 5)
#     assert len(db.addData.groupby('data.y.y_2_id').first()['data.y.y_2'])==len(time_points_answ), \
#         'scale_data_with_nan_rows: wrong number of points'
#     assert all([abs(t-answ)<0.01 for t, answ in zip(db.addData.groupby('data.y.y_2_id').first()['data.y.y_2'], time_points_answ)]), \
#         'scale_data_with_nan_rows: wrong points centers'

def test_get_statistics():
    df = citros.topic('A').rid(start = 0, end = 10).data(['data.x', 'data.time', 'data.p'])
    dataset = CitrosData(df, type_name = 'x', data_label=['data.x'], units = 'm')
    db_sc = dataset.scale_data(n_points = 5, 
                                    param_label = 'data.time', 
                                    show_fig = False)

    def check_statistic_values(time_id):
        assert (db_sc.get_statistics().loc[time_id]['mean']-np.array(db_sc.data.loc[time_id].mean())).sum() < 0.01,\
            'mean value for time_id = '+str(time_id)+' is wrong'
        assert (np.array(db_sc.data.loc[time_id].cov())-db_sc.get_statistics().loc[time_id]['covar_matrix']).sum() < 0.01,\
            'covariant matrix for time_id = '+str(time_id)+' is wrong'
        assert (np.sqrt(np.diag(np.array(db_sc.data.loc[time_id].cov())))-db_sc.get_statistics().loc[time_id]['sigma']).sum() < 0.01,\
            'standard deviation for time_id = '+str(time_id)+' is wrong'
    
    check_statistic_values(0)
    check_statistic_values(1)
    check_statistic_values(2)
    check_statistic_values(3)
    check_statistic_values(4)

    df = citros.topic('A').data(['data.y.y_1', 'data.y.y_2'])
    #get_statistics with nan and inf, 1 col
    #bin
    dataset = CitrosData(df, data_label = 'data.y.y_1', units = 'm')
    db = dataset.bin_data(n_bins = 10, param_label = 'data.y.y_2', show_fig = False)
    stat = db.get_statistics(return_format='citrosStat')
    assert (db.data.groupby(db.xid_label).mean() == stat.mean)['data.y.y_1'].all(), 'get_statistics with nan and inf, bin: mean is wrong'
    assert (stat.covar_matrix.apply(lambda x: x[0][0]).fillna(True) == 
            db.data.groupby(db.xid_label).cov().groupby(level = 0).first()['data.y.y_1'].fillna(True)).all(), \
            'get_statistics with nan and inf, bin: cov is wrong'
    assert (db.data.groupby(db.xid_label).std().fillna(True) == stat.std.fillna(True))['data.y.y_1'].all(), \
            'get_statistics with nan and inf, bin: std is wrong'
    #scale
    db = dataset.scale_data(n_points = 10, param_label = 'data.y.y_2', show_fig = False)
    stat = db.get_statistics(return_format='citrosStat')
    assert (db.data.groupby(db.xid_label).mean() == stat.mean)['data.y.y_1'].all(), 'get_statistics with nan and inf, scale: wrong'
    assert (stat.covar_matrix.apply(lambda x: x[0][0]).fillna(True) == 
            db.data.groupby(db.xid_label).cov().groupby(level = 0).first()['data.y.y_1'].fillna(True)).all(), \
            'get_statistics with nan and inf, scale: cov is wrong'
    assert (db.data.groupby(db.xid_label).std().fillna(True) == stat.std.fillna(True))['data.y.y_1'].all(), \
            'get_statistics with nan and inf, scale: std is wrong'
    
    df = citros.topic('A').data(['data.y.y_1', 'data.y.y_2', 'data.y.y_3'])
    #get_statistics with nan and inf 2 cols, omit_nan_rows = True
    #bin
    dataset = CitrosData(df, data_label = ['data.y.y_1', 'data.y.y_3'], units = 'm')
    db = dataset.bin_data(n_bins = 10, param_label = 'data.y.y_2', show_fig = False)
    stat = db.get_statistics(return_format='citrosStat')
    assert (db.data.groupby(db.xid_label).mean() == stat.mean).all().all(),\
            'get_statistics with nan and inf 2cols, bin: mean is wrong'
    
    assert (db.data.groupby(db.xid_label).cov().xs('data.y.y_1', level = 1)['data.y.y_1'] == 
            stat.covar_matrix.apply(lambda x: x[0][0])).all() and \
            (db.data.groupby(db.xid_label).cov().xs('data.y.y_1', level = 1)['data.y.y_3'] == 
            stat.covar_matrix.apply(lambda x: x[0][1])).all() and \
            (db.data.groupby(db.xid_label).cov().xs('data.y.y_3', level = 1)['data.y.y_3'] == 
            stat.covar_matrix.apply(lambda x: x[1][1])).all(), \
            'get_statistics with nan and inf, bin 2cols: cov is wrong'
    
    assert (db.data.groupby(db.xid_label).std() == stat.std).all().all(), \
            'get_statistics with nan and inf, bin 2cols: std is wrong'
    #scale
    db = dataset.scale_data(n_points = 10, param_label = 'data.y.y_2', show_fig = False)
    stat = db.get_statistics(return_format='citrosStat')
    assert (db.data.groupby(db.xid_label).mean().fillna(True) == stat.mean.fillna(True)).all().all(),\
            'get_statistics with nan and inf, scale 2cols: mean is wrong'
    assert (db.data.groupby(db.xid_label).cov().fillna(True).xs('data.y.y_1', level = 1)['data.y.y_1'] == 
            stat.covar_matrix.apply(lambda x: x[0][0]).fillna(True)).all() and \
            (db.data.groupby(db.xid_label).cov().fillna(True).xs('data.y.y_1', level = 1)['data.y.y_3'] == 
            stat.covar_matrix.apply(lambda x: x[0][1]).fillna(True)).all() and \
            (db.data.groupby(db.xid_label).cov().fillna(True).xs('data.y.y_3', level = 1)['data.y.y_3'] == 
            stat.covar_matrix.apply(lambda x: x[1][1]).fillna(True)).all(), \
            'get_statistics with nan and inf, scale 2cols: cov is wrong'
    assert (db.data.groupby(db.xid_label).std().fillna(True) == stat.std.fillna(True)).all().all(), \
            'get_statistics with nan and inf, scale 2cols: std is wrong'
    
    #get_statistics with nan and inf, omit_nan_rows = False
    dataset = CitrosData(df, data_label = ['data.y.y_1', 'data.y.y_3'], units = 'm', omit_nan_rows = False)
    db = dataset.bin_data(n_bins = 10, param_label = 'data.y.y_2', show_fig = False)
    stat = db.get_statistics(return_format='citrosStat')
    assert (db.data.groupby(db.xid_label).mean() == stat.mean).all().all(),\
            'get_statistics with nan and inf 2cols, bin, omit_nan_rows = False: mean is wrong'
    assert (db.data.groupby(db.xid_label).cov().xs('data.y.y_1', level = 1)['data.y.y_1'] == 
            stat.covar_matrix.apply(lambda x: x[0][0])).all() and\
            ((db.data.groupby(db.xid_label).cov().xs('data.y.y_1', level = 1)['data.y.y_3'] == 
            stat.covar_matrix.apply(lambda x: x[0][1])).all()) and \
            ((db.data.groupby(db.xid_label).cov().xs('data.y.y_3', level = 1)['data.y.y_3'] == 
            stat.covar_matrix.apply(lambda x: x[1][1])).all()), \
            'get_statistics with nan and inf 2cols, bin 2cols, omit_nan_rows = False: cov is wrong'
    assert (db.data.groupby(db.xid_label).std() == stat.std).all().all(), \
            'get_statistics with nan and inf, bin 2cols, omit_nan_rows = False: std is wrong'

    db = dataset.scale_data(n_points = 10, param_label = 'data.y.y_2', show_fig = False)
    stat = db.get_statistics(return_format='citrosStat')
    assert (db.data.groupby(db.xid_label).mean().fillna(True) == stat.mean.fillna(True)).all().all(),\
            'get_statistics with nan and inf, scale 2cols, omit_nan_rows = False: mean is wrong'
    assert (db.data.groupby(db.xid_label).cov().fillna(True).xs('data.y.y_1', level = 1)['data.y.y_1'] == 
            stat.covar_matrix.apply(lambda x: x[0][0]).fillna(True)).all() and \
            (db.data.groupby(db.xid_label).cov().fillna(True).xs('data.y.y_1', level = 1)['data.y.y_3'] == 
            stat.covar_matrix.apply(lambda x: x[0][1]).fillna(True)).all() and \
            (db.data.groupby(db.xid_label).cov().fillna(True).xs('data.y.y_3', level = 1)['data.y.y_3'] == 
            stat.covar_matrix.apply(lambda x: x[1][1]).fillna(True)).all(), \
            'get_statistics with nan and inf, scale 2cols, omit_nan_rows = False: cov is wrong'
    assert (db.data.groupby(db.xid_label).std().fillna(True) == stat.std.fillna(True)).all().all(), \
            'get_statistics with nan and inf, scale 2cols, omit_nan_rows = False: std is wrong'

def test_show_statistics():
    plt.close('all')
    df = citros.topic('C').data(['data.x.x_1','data.x.x_2', 'data.time', 'data.p'])
    dataset = CitrosData(df, type_name = 'x', data_label=['data.x.x_1', 'data.x.x_2'], units = 'm')
    db_sc = dataset.scale_data(n_points = 35, 
                            param_label = 'data.time', 
                            show_fig = False)
    fig, axes = db_sc.show_statistics(return_fig=True)

    for ax in axes:
        n_lines = len(ax.get_lines())
        assert n_lines == 6, 'show_statistics: number of lines is wrong'

def test_show_statistics_1col_nan_rows_false():
    plt.close('all')
    df = citros.topic('A').data(['data.y.y_1', 'data.y.y_2', 'data.time'])
    dataset = CitrosData(df, data_label = 'data.y.y_1', units = 'm', omit_nan_rows = False)
    db = dataset.scale_data(n_points = 10, param_label = 'data.y.y_2', show_fig = False)
    fig, axes = db.show_statistics(std_color = 'b', return_fig = True)
    for ax in axes:
        n_lines = len(ax.get_lines())
        assert n_lines == 6, 'show_statistics_1col_nan_rows_false: number of lines is wrong'

def test_show_statistics_1col_nan_rows_false():
    plt.close('all')
    df = citros.topic('A').data(['data.y.y_1', 'data.y.y_2', 'data.time'])
    dataset = CitrosData(df, data_label = 'data.y.y_1', units = 'm', omit_nan_rows = False)
    db = dataset.scale_data(n_points = 10, param_label = 'data.y.y_2', show_fig = False)
    fig, axes = db.show_statistics(std_color = 'b', return_fig = True, connect_nan_std = False, std_area = True, std_lines = True)

    for ax in axes:
        n_lines = len(ax.get_lines())
        assert n_lines == 6, 'show_statistics_1col_nan_rows_false: number of lines is wrong'

def test_show_statistics_1col_nan_rows_true():
    plt.close('all')
    df = citros.topic('A').data(['data.y.y_1', 'data.y.y_2', 'data.time'])
    dataset = CitrosData(df, data_label = 'data.y.y_1', units = 'm', omit_nan_rows = True)
    db = dataset.bin_data(n_bins = 50, param_label = 'data.y.y_2', show_fig = False)
    fig, axes = db.show_statistics(return_fig = True, connect_nan_std = False, std_area = True, std_lines = False)
    for ax in axes:
        n_lines = len(ax.get_lines())
        assert n_lines == 4, 'show_statistics_1col_nan_rows_true: number of lines is wrong'

def test_show_statistics_2col_nan_rows_true():
    plt.close('all')
    df = citros.topic('A').data(['data.y.y_3', 'data.y.y_1', 'data.y.y_2', 'data.time'])
    dataset = CitrosData(df, data_label = ['data.y.y_1', 'data.y.y_3'], units = 'm', omit_nan_rows = True)
    db = dataset.scale_data(n_points = 10, param_label = 'data.y.y_2', show_fig = False)
    fig, axes = db.show_statistics(return_fig = True, connect_nan_std = True, std_area = True, std_lines = False)
    for ax in axes:
        n_lines = len(ax.get_lines())
        assert n_lines == 4, 'show_statistics_2col_nan_rows_true: number of lines is wrong'

def test_show_statistics_2col_nan_rows_false():
    plt.close('all')
    df = citros.topic('A').data(['data.y.y_3', 'data.y.y_1', 'data.y.y_2', 'data.time'])
    dataset = CitrosData(df, data_label = ['data.y.y_1', 'data.y.y_3'], units = 'm', omit_nan_rows = False)
    db = dataset.bin_data(n_bins = 10, param_label = 'data.y.y_2', show_fig = False)
    fig, axes = db.show_statistics(return_fig = True)
    for ax in axes:
        n_lines = len(ax.get_lines())
        assert n_lines == 6, 'show_statistics_2col_nan_rows_false: number of lines is wrong'

def test_show_statistics_1col():
    plt.close('all')
    fig = plt.figure(figsize = (6,6))
    df = citros.topic('C').data(['data.x.x_1', 'data.time', 'data.p'])
    dataset = CitrosData(df, type_name = 'x', data_label=['data.x.x_1'], units = 'm')
    #dataset.data = dataset.data.add(p)
    db_sc = dataset.scale_data(n_points = 35, 
                            param_label = 'data.time', 
                            show_fig = False)
    fig, axes = db_sc.show_statistics(fig =fig, return_fig=True)
    for ax in axes:
        n_lines = len(ax.get_lines())
        assert n_lines == 6, 'show_statistics_1col: number of lines is wrong'
    
def test_show_correlation_output():
    #when passing figure to plot
    plt.close('all')
    fig = plt.figure(figsize = (6,6))
    df = citros.topic('C').data(['data.x.x_1','data.x.x_2', 'data.time', 'data.p'])
    dataset = CitrosData(df, type_name = 'x', data_label=['data.x.x_1', 'data.x.x_2'], units = 'm')
    #dataset.data = dataset.data.add(p)
    db_sc = dataset.scale_data(n_points = 35, 
                            param_label = 'data.time', 
                            show_fig = False)
    db_sc.show_correlation(x_col = 'data.x.x_1',
                        y_col = 'data.x.x_2',
                        slice_id = 2,
                        n_std = [1,2,3], display_id=False, fig = fig)
    
    for ax in fig.axes:
        n_lines = len(ax.get_lines())
        n_patches = len(ax.patches)
        assert n_lines == 2, 'show_correlation_output: number of lines is wrong'
        assert n_patches == 3, 'show_correlation_output: number of patches is wrong'

def test_show_correlation_output_2():
    #when creating figure to plot and return it
    plt.close('all')
    df = citros.topic('C').data(['data.x.x_1','data.x.x_2', 'data.time', 'data.p'])
    dataset = CitrosData(df, type_name = 'x', data_label=['data.x.x_1', 'data.x.x_2'], units = 'm')
    db_sc = dataset.scale_data(n_points = 35, 
                            param_label = 'data.time', 
                            show_fig = False)
    fig, ax = db_sc.show_correlation(x_col = 'data.x.x_1',
                                       y_col = 'data.x.x_2',
                                       slice_id = 2,
                                       n_std = [1,2,3], display_id=False, return_fig = True)
    
    n_lines = len(ax.get_lines())
    n_patches = len(ax.patches)
    assert n_lines == 2, 'show_correlation_output_2: number of lines is wrong'
    assert n_patches == 3, 'show_correlation_output_2: number of patches is wrong'

def test_show_correlation_nan_inf():
    df = citros.topic('A').data(['data.y.y_1', 'data.y.y_2', 'data.y.y_3', 'data.time'])
    n_lines_answ_sc = [0, 4, 4, 4, 4, 4, 4, 4, 4, 1]
    n_patches_answ_sc = [0, 1, 1, 1, 1, 1, 1, 1, 1, 0]

    n_lines_answ_bin = [4]*10
    n_patches_answ_bin = [1]*10

    dataset = CitrosData(df, data_label = ['data.y.y_2', 'data.y.y_3'], units = 'm', omit_nan_rows = True)
    db = dataset.scale_data(n_points = 10, param_label = 'data.y.y_1', show_fig = False)
    for i in list(set(db.data.index.get_level_values('data.y.y_1_id'))):
        fig, ax = db.show_correlation(x_col = 'data.y.y_2', y_col = 'data.y.y_3', slice_id = i, bounding_error=True, display_id=False, return_fig = True)
        assert len(ax.get_lines()) == n_lines_answ_sc[i], f'show_correlation_nan_inf: the number of lines for slice_id = {i} is incorrect'
        assert len(ax.patches) == n_patches_answ_sc[i], f'show_correlation_nan_inf: the number of patches for slice_id = {i} is incorrect'
    plt.close('all')

    db = dataset.bin_data(n_bins = 10, param_label = 'data.y.y_1', show_fig = False)
    for i in list(set(db.data.index.get_level_values('data.y.y_1_id'))):
        fig, ax = db.show_correlation(x_col = 'data.y.y_2', y_col = 'data.y.y_3', slice_id = i, bounding_error=True, display_id=False, n_std = 1, return_fig =True)
        assert len(ax.get_lines()) == n_lines_answ_bin[i], f'the number of lines for slice_id = {i} is incorrect'
        assert len(ax.patches) == n_patches_answ_bin[i], f'the number of patches for slice_id = {i} is incorrect'
    plt.close('all')

    dataset = CitrosData(df, data_label = ['data.y.y_2', 'data.y.y_3'], units = 'm', omit_nan_rows = False)
    db = dataset.scale_data(n_points = 10, param_label = 'data.y.y_1', show_fig = False)
    for i in list(set(db.data.index.get_level_values('data.y.y_1_id'))):
        fig, ax = db.show_correlation(x_col = 'data.y.y_2', y_col = 'data.y.y_3', slice_id = i, bounding_error=True, display_id=False, return_fig = True)
        assert len(ax.get_lines()) == n_lines_answ_sc[i], f'show_correlation_nan_inf: the number of lines for slice_id = {i} is incorrect'
        assert len(ax.patches) == n_patches_answ_sc[i], f'show_correlation_nan_inf: the number of patches for slice_id = {i} is incorrect'
    plt.close('all')

    db = dataset.bin_data(n_bins = 10, param_label = 'data.y.y_1', show_fig = False)
    for i in list(set(db.data.index.get_level_values('data.y.y_1_id'))):
        fig, ax = db.show_correlation(x_col = 'data.y.y_2', y_col = 'data.y.y_3', slice_id = i, bounding_error=True, display_id=False, n_std = 1, return_fig =True)
        assert len(ax.get_lines()) == n_lines_answ_bin[i], f'the number of lines for slice_id = {i} is incorrect'
        assert len(ax.patches) == n_patches_answ_bin[i], f'the number of patches for slice_id = {i} is incorrect'
    plt.close('all')

def test_show_correlation_output_2db():
    #when two databases
    plt.close('all')
    df = citros.topic('C').data(['data.x.x_1','data.x.x_2', 'data.time', 'data.p'])
    dataset = CitrosData(df, type_name = 'x', data_label=['data.x.x_1', 'data.x.x_2'], units = 'm')
    #dataset.data = dataset.data.add(p)
    db_sc = dataset.scale_data(n_points = 35, 
                            param_label = 'data.time', 
                            show_fig = False)
    db_sc2 = dataset.scale_data(n_points = 60, 
                            param_label = 'data.time', 
                            show_fig = False)
    _, ax = db_sc.show_correlation(db2 = db_sc2,
                           x_col = 'data.x.x_1',
                           y_col = 'data.x.x_2',
                           slice_id = 2,
                           n_std = 1, display_id=False, return_fig=True)

    n_lines = len(ax.get_lines())
    n_patches = len(ax.patches)
    assert n_lines == 2, 'show_correlation_output_2: number of lines is wrong'
    assert n_patches == 1, 'show_correlation_output_2: number of patches is wrong'

def test_show_correlation_output_bound_error():
    plt.close('all')
    fig = plt.figure(figsize = (6,6))
    df = citros.topic('C').data(['data.x.x_1','data.x.x_2', 'data.time', 'data.p'])
    dataset = CitrosData(df, type_name = 'x', data_label=['data.x.x_1', 'data.x.x_2'], units = 'm')
    #dataset.data = dataset.data.add(p)
    db_sc = dataset.scale_data(n_points = 35, 
                            param_label = 'data.time', 
                            show_fig = False)
    db_sc.show_correlation(x_col = 'data.x.x_1',
                        y_col = 'data.x.x_2',
                        slice_id = 2,
                        n_std = [1,2], display_id=False, fig = fig, bounding_error=True)
    for ax in fig.axes:
        n_lines = len(ax.get_lines())
        n_patches = len(ax.patches)
    assert n_lines == 5, 'show_correlation_output_bound_error: number of lines is wrong'
    assert n_patches == 2, 'show_correlation_output_bound_error: number of patches is wrong'

def test_set_parameter():
    #I parameter setting by parameter_label
    df = citros.topic('A').set_filter({'data.t':[10]}).data(['data.x.x_1', 'data.time', 'data.p', 'data.t'])
    dataset = CitrosData(df, type_name = 'x', data_label=['data.x.x_1'], units = 'm', parameter_label='data.t')
    assert abs(dataset.parameters['data.t'] - 10) < 0.001, 'I parameter setting by parameter_label is wrong'

    #II parameter setting by parameter_label
    df = citros.topic('A').set_filter({'data.t':[10]}).data(['data.x.x_1', 'data.time', 'data.p', 'data.t'])
    dataset = CitrosData(df, type_name = 'x', data_label=['data.x.x_1'], units = 'm', parameter_label=['data.t', 'data.p'])
    assert abs(dataset.parameters['data.t'] - 10) < 0.001, 'II parameter setting by parameter_label is wrong'
    assert abs(dataset.parameters['data.p'] - 0.8) < 0.001, 'II parameter setting by parameter_label is wrong'

    #III parameter setting parameters dict
    df = citros.topic('A').set_filter({'data.t':[10]}).data(['data.x.x_1', 'data.time', 'data.p', 'data.t'])
    dataset = CitrosData(df, type_name = 'x', data_label=['data.x.x_1'], units = 'm', parameters = {'m': 55, 'k': 94})
    assert abs(dataset.parameters['m']-55) < 0.001, 'III parameter setting by parameters dict is wrong'
    assert abs(dataset.parameters['k']-94) < 0.001, 'III parameter setting by parameters dict is wrong'

    #IV parameter setting by set_parameter
    dataset.set_parameter(key = 'v', value = 10)
    dataset.set_parameter(item = {'f': 5, 'g': 1.5})
    assert abs(dataset.parameters['v'] - 10) < 0.001, 'IV parameter setting by set_parameter is wrong'
    assert abs(dataset.parameters['f'] - 5) < 0.001, 'IV parameter setting by set_parameter is wrong'
    assert abs(dataset.parameters['g'] - 1.5) < 0.001, 'IV parameter setting by set_parameter is wrong'
    dataset.drop_parameter('v')
    assert 'v' not in dataset.parameters.keys(), 'IV parameter setting, drop_parameter works wrong'

def test_prediction():
    db_array = CitrosDataArray()
    parameters = [-0.2, 0, 0.2, 0.6]
    
    df = citros.topic('C').data(['data.x.x_1', 'data.time', 'data.p'])
    for p in parameters:
        dataset = CitrosData(df, type_name = 'x', data_label=['data.x.x_1'], units = 'm')
        dataset.set_parameter('p', p)
        dataset.data = dataset.data.add(p)
        db_sc = dataset.scale_data(n_points = 35, 
                                param_label = 'data.time', 
                                show_fig = False)
        db_array.add_db(db_sc)
    p_pred = 0.4
    predicted = db_array.get_prediction(parameters = {'p': p_pred}, 
                                        method = ['neural_net', 'poly','gmm'], 
                                        n_poly = 2, 
                                        activation='tanh', solver='lbfgs', hidden_layer_sizes = (30,), random_state = 9)

    dataset = CitrosData(df, type_name = 'x', data_label=['data.x.x_1'], units = 'm')
    dataset.data = dataset.data.add(p_pred)
    db_sc = dataset.scale_data(n_points = 35, 
                            param_label = 'data.time', 
                            show_fig = False)
    stat = db_sc.get_statistics()

    assert all((predicted[0]['data.x.x_1']-stat['mean'].apply(lambda x: x[0]))/predicted[0]['data.x.x_1']<0.1),\
        'get_prediction(): "neural_net", difference more than 10 percents'
    assert all((predicted[1]['data.x.x_1']-stat['mean'].apply(lambda x: x[0]))/predicted[0]['data.x.x_1']<0.1),\
        'get_prediction(): "poly", difference more than 10 percents'
    assert all((predicted[2]['data.x.x_1']-stat['mean'].apply(lambda x: x[0]))/predicted[0]['data.x.x_1']<0.1),\
        'get_prediction(): "gmm", difference more than 10 percents'

def test_show_prediction():
    plt.close('all')
    db_array = CitrosDataArray()
    parameters = [-0.2, 0, 0.2, 0.6]
    
    df = citros.topic('C').data(['data.x.x_1', 'data.time', 'data.p'])
    for p in parameters:
        dataset = CitrosData(df, type_name = 'x', data_label=['data.x.x_1'], units = 'm')
        dataset.set_parameter('p', p)
        dataset.data = dataset.data.add(p)
        db_sc = dataset.scale_data(n_points = 35, 
                                param_label = 'data.time', 
                                show_fig = False)
        db_array.add_db(db_sc)

    p_pred = 0.4
    _, fig, ax = db_array.get_prediction(parameters = {'p': p_pred}, 
                                        method = ['neural_net', 'poly','gmm'], 
                                        n_poly = 2, 
                                        activation='tanh', solver='lbfgs', hidden_layer_sizes = (30,), random_state = 9,  
                                        show_fig = True, return_fig=True)
    n_lines = len(ax.get_lines())
    assert n_lines == 7, 'show_prediction: number of lines is wrong'
    
def test_show_prediction_2par():
    plt.close('all')
    fig = plt.figure(figsize = (6,6))
    db_array = CitrosDataArray()
    parameters = [-0.2, 0, 0.2, 0.6]
    
    df = citros.topic('C').data(['data.x.x_1', 'data.time', 'data.p'])
    for p in parameters:
        dataset = CitrosData(df, type_name = 'x', data_label=['data.x.x_1'], units = 'm')
        dataset.set_parameter('p', p)
        dataset.set_parameter('t', p)
        dataset.data = dataset.data.add(p)
        db_sc = dataset.scale_data(n_points = 35, 
                                param_label = 'data.time', 
                                show_fig = False)
        db_array.add_db(db_sc)
    p_pred = 0.4
    _ = db_array.get_prediction(parameters = {'p': p_pred, 't': p_pred},
                                        method = ['neural_net', 'poly','gmm'], 
                                        n_poly = 2, 
                                        activation='tanh', solver='lbfgs', hidden_layer_sizes = (30,), random_state = 9,  
                                        fig = fig)
    n_lines = len(fig.axes[0].get_lines())
    assert n_lines == 7, 'show_prediction: number of lines is wrong'
    
def test_show_prediction_3col():
    plt.close('all')
    fig = plt.figure(figsize = (6,6))
    db_array = CitrosDataArray()
    parameters = [-0.2, 0, 0.2, 0.6]
    
    df = citros.topic('C').data(['data.x.x_1', 'data.x.x_2', 'data.time', 'data.p'])
    for p in parameters:
        dataset = CitrosData(df, type_name = 'x', data_label=['data.x.x_1', 'data.x.x_2'], units = 'm')
        dataset.set_parameter('p', p)
        dataset.data = dataset.data.add(p)
        db_sc = dataset.scale_data(n_points = 35, 
                                param_label = 'data.time', 
                                show_fig = False)
        db_array.add_db(db_sc)

    p_pred = 0.4
    _ = db_array.get_prediction(parameters = {'p': p_pred}, 
                                        method = ['neural_net', 'poly','gmm'], 
                                        n_poly = 2, 
                                        activation='tanh', solver='lbfgs', hidden_layer_sizes = (30,), random_state = 9,  
                                        fig = fig)
    
    for ax in fig.axes:
        n_lines = len(ax.get_lines())
        assert n_lines == 7, 'show_prediction: number of lines is wrong'


def test_get_prediction_with_nan_inf():
    
    def compare(omit_nan_rows, method):
        list_t = [10,20,40]

        db_array = CitrosDataArray()
        for t in list_t:
            df = citros.topic('A').\
                        data(['data.y.y_1', 'data.y.y_2', 'data.y.y_3']).assign(**{'data.y.y_1': lambda F: F['data.y.y_1']+t,
                                                                    'data.y.y_2': lambda F: F['data.y.y_2']+t})
            dataset = CitrosData(df, data_label=['data.y.y_1', 'data.y.y_2'], units = 'm', omit_nan_rows = omit_nan_rows )
            dataset.set_parameter('t', t)
            if method == 'scale':
                db_sc = dataset.scale_data(n_points = 10, 
                                                param_label = 'data.y.y_3', 
                                                show_fig = False)
            elif method == 'bin': 
                db_sc = dataset.bin_data(n_bins = 10, 
                                                param_label = 'data.y.y_3', 
                                                show_fig = False)
            db_array.add_db(db_sc)

        df = citros.topic('A').\
                    data(['data.y.y_1', 'data.y.y_2', 'data.y.y_3']).assign(**{'data.y.y_1': lambda F: F['data.y.y_1']+30,
                                                                    'data.y.y_2': lambda F: F['data.y.y_2']+30})
        dataset = CitrosData(df, data_label=['data.y.y_1', 'data.y.y_2'], units = 'm', omit_nan_rows = omit_nan_rows )
        dataset.set_parameter('t', t)
        if method == 'scale':
            db_answ = dataset.scale_data(n_points = 10, 
                                            param_label = 'data.y.y_3', 
                                            show_fig = False)
        elif method == 'bin': 
            db_answ = dataset.bin_data(n_bins = 10, 
                                            param_label = 'data.y.y_3', 
                                            show_fig = False)

        fig = plt.figure(figsize = (8,8))
        prediction_method = ['neural_net', 'poly','gmm']
        predicted = db_array.get_prediction(parameters = {'t':30}, 
                                            method = prediction_method, 
                                            n_poly = 2, 
                                            activation='tanh', solver='lbfgs', hidden_layer_sizes = (15,), random_state = 9,  
                                            show_fig = False,
                                            fig = fig)

        limit_list = [0.2, 1e-5, 1e-5]
        for i, limit in enumerate(limit_list):
            print(f'{omit_nan_rows}')
            print(f'{method}')
            print(predicted[i][['data.y.y_1', 'data.y.y_2']])
            print(db_answ.get_statistics(return_format = 'citrosStat').mean)
            print((predicted[i][['data.y.y_1', 'data.y.y_2']] - db_answ.get_statistics(return_format = 'citrosStat').mean).fillna(0).abs() < limit )
            assert ((predicted[i][['data.y.y_1', 'data.y.y_2']] - db_answ.get_statistics(return_format = 'citrosStat').mean).fillna(0).abs() < limit).all().all() == True,\
                f"test_get_prediction_with_nan_inf: omit_nan_rows = {omit_nan_rows}, method = {method}, prediction method = {prediction_method[i]}: result is wrong"

    compare(omit_nan_rows = True, method = 'bin')
    compare(omit_nan_rows = True, method = 'scale')
    compare(omit_nan_rows = False, method = 'bin')
    compare(omit_nan_rows = False, method = 'scale')



import os
import re
import git
import ast
import glob
import yaml
import json
from rich import inspect
from pathlib import Path

import xml.etree.ElementTree as ET

from .parser_base import ParserBase


class ParserRos2(ParserBase):
    """
    Class to parse ROS2 package files (XML, CMakeLists.txt, setup.py etc).
    """

    def __init__(self, log, ignore_list=[]):
        self.log = log
        self.project = None
        self.ignore_list = ignore_list

    ################################ Any lang #################################

    def parse_xml(self, package_path):
        """
        Parse an XML file in the given package path.

        :param package_path: Path of the package.
        """
        path_to_package_xml = Path(package_path, "package.xml")
        if not path_to_package_xml.exists():
            self.log.error(f"File not found: {str(path_to_package_xml)}")
            return {}

        path_to_package_xml = str(path_to_package_xml)

        try:
            tree = ET.parse(path_to_package_xml)
            root = tree.getroot()
        except ET.ParseError as ex:
            self.log.exception(
                f"parsing error while trying to parse: {path_to_package_xml}"
            )
            return {}

        package_name = root.find("name")
        version = root.find("version")
        maintainer = root.find("maintainer")
        maintainer_email = maintainer.attrib["email"] if maintainer is not None else ""
        description = root.find("description")

        export = root.find("export")
        build_type = export.find("build_type") if export is not None else None

        return {
            "package_xml": path_to_package_xml,
            "package_name": package_name.text if package_name is not None else "",
            "version": version.text if version is not None else "",
            "maintainer": maintainer.text if maintainer is not None else "",
            "maintainer_email": maintainer_email,
            "description": description.text if description is not None else "",
            "nodes": [],
            "build_type": build_type.text if build_type is not None else None,
        }

    ################################# C / CPP #################################

    def parse_makefile(self, package_path):
        path_to_cmake = Path(package_path, "CMakeLists.txt")

        try:
            nodes = self.find_install_targets(path_to_cmake)
        except Exception as e:
            self.log.exception(
                f"Exception raised while trying to parse {path_to_cmake}"
            )
            print(f"[red]Could not parse {path_to_cmake}[/red]")
            nodes = []

        return {"cmake": str(path_to_cmake), "nodes": nodes}

    def find_install_targets(self, cmake_file_path: Path):
        """
        Parses the given CMakelists.txt file and returns a list of nodes,
        whose names are the install targets defined in the file.

        Unhandled edge cases:
        1.  Multiple assignments: If a variable is set multiple times in the
            CMakeLists.txt file, the value that this function will use is the
            value that was set last before the `install` command.
        2.  Conditional assignments: If a variable is set inside an if statement,
            the function will still process it regardless of the condition.
        3.  Function definitions: If `install` is called inside a function definition,
            and a target variable used is a function parameter, an exception will be
            raised (since the variable will not be recognized).
        4.  Nested variable references (e.g., ${${ANOTHER_VAR}}).
        """
        # import locally to avoid conflicts with the `ast` (for python) module.
        from cmakeast import ast
        from cmakeast.ast_visitor import recurse

        with open(cmake_file_path, "r") as f:
            contents = f.read()

        # Parse the CMake file to an AST
        root = ast.parse(contents)

        var_assignments = {"PROJECT_NAME": str(cmake_file_path.parent.name)}

        # List to store all install targets
        install_target_nodes = []

        def add_node(node_name):
            install_target_nodes.append(
                {"name": node_name, "entry_point": "", "path": "", "parameters": []}
            )
            # print(f"     node: {node_name}")

        def unquote(string):
            if string.startswith('"') and string.endswith('"'):
                return string[1:-1]
            return string

        def get_var_assignments(set_node):
            if len(set_node.arguments) >= 2:
                # The first argument is the variable name
                var_name = set_node.arguments[0].contents

                # Concatenate all subsequent arguments assuming they may be part of the list
                values_list = set_node.arguments[1:]

                # CMake lists can be separated by whitespace or semicolons.
                value = ";".join(arg.contents for arg in values_list)

                # Remove surrounding quotes from the entire value if present
                value = unquote(value)

                # We split on whitespace and flatten the result in case any individual
                # element was already a semicolon-separated list.
                # Split on whitespace first
                value_split = value.split()

                # Process each element by splitting on semicolons
                sublists = (elem.split(";") for elem in value_split)

                # Flatten the list of lists, remove individual quotes, and filter out empty strings
                value_elements = [
                    unquote(item) for sublist in sublists for item in sublist if item
                ]
                var_assignments[var_name] = value_elements

        validity_checks = {
            "exec_found": False,
            "install_found": False,
            "at_least_one_node": False,
        }

        # Known properties - see https://cmake.org/cmake/help/latest/command/install.html#targets
        properties = {
            "EXPORT",
            "RUNTIME_DEPENDENCIES",
            "RUNTIME_DEPENDENCY_SET",
            "ARCHIVE",
            "LIBRARY",
            "RUNTIME",
            "OBJECTS",
            "FRAMEWORK",
            "BUNDLE",
            "PRIVATE_HEADER",
            "PUBLIC_HEADER",
            "RESOURCE",
            "FILE_SET",
            "CXX_MODULES_BMI",
            "DESTINATION",
            "PERMISSIONS",
            "CONFIGURATIONS",
            "COMPONENT",
            "NAMELINK_COMPONENT",
            "OPTIONAL",
            "EXCLUDE_FROM_ALL",
            "NAMELINK_ONLY",
            "NAMELINK_SKIP",
            "INCLUDES DESTINATION",
        }

        cmake_file_path = str(cmake_file_path)

        # Define a handler for FunctionCall nodes
        def handle_function_call(name, node, depth):
            # Handle `set` commands to track variable assignments
            if node.name == "set":
                get_var_assignments(node)
            # Handle install commands to track install targets
            elif node.name == "add_executable":
                # If there is an executable, there should be an install
                validity_checks["exec_found"] = True
            elif node.name == "install":
                validity_checks["install_found"] = True
                arguments = node.arguments
                if arguments and arguments[0].contents == "TARGETS":
                    # Loop through the arguments until a property is found
                    for arg in arguments[1:]:
                        if arg.contents in properties:
                            break
                        validity_checks["at_least_one_node"] = True
                        if arg.contents.startswith("${") and arg.contents.endswith("}"):
                            var_name = arg.contents[2:-1]
                            if var_name in var_assignments:
                                var_value = var_assignments[var_name]
                                if isinstance(var_value, list):
                                    for target in var_value:
                                        add_node(target)
                                elif isinstance(var_value, str):
                                    add_node(var_value)
                            else:
                                raise ValueError(
                                    f"Unknown cmake variable name: {var_name} in {cmake_file_path}"
                                )
                        else:
                            add_node(arg.contents)

        # Use the recurse function to traverse the AST with the handler
        recurse(root, function_call=handle_function_call)

        if not validity_checks["install_found"] and not validity_checks["exec_found"]:
            self.log.debug(
                f"{cmake_file_path} is a pure interface package without install and nodes."
            )
        else:
            if not validity_checks["install_found"]:
                self.log.error(
                    f"{cmake_file_path} is not formatted correctly: `add_executable` with no 'install' command."
                )
                raise ValueError(
                    f"{cmake_file_path} is not formatted correctly: `add_executable` with no 'install' command."
                )
            if not validity_checks["at_least_one_node"]:
                self.log.debug(f"{cmake_file_path} does not install any nodes.")

        return install_target_nodes

    ################################# Python ##################################

    def _extract_contents(self, node, global_scope):
        """
        Recursive helper function for parsing a variety of objects such as lists, function calls etc.
        Does not handle every case, such as nested functions etc. Returns None on failure.
        """
        if isinstance(node, ast.Str):
            return node.s
        elif isinstance(node, ast.Num):
            return node.n
        elif isinstance(node, ast.Name):
            return global_scope.get(node.id)
        elif isinstance(node, ast.List):
            return [self._extract_contents(item, global_scope) for item in node.elts]
        elif isinstance(node, ast.Tuple):
            return tuple(
                self._extract_contents(item, global_scope) for item in node.elts
            )
        elif isinstance(node, ast.Dict):
            return {
                self._extract_contents(key, global_scope): self._extract_contents(
                    value, global_scope
                )
                for key, value in zip(node.keys, node.values)
            }
        elif isinstance(node, ast.BinOp):
            return self._extract_contents(
                node.left, global_scope
            ) + self._extract_contents(node.right, global_scope)
        elif isinstance(node, ast.Call):
            pos_args = tuple(
                self._extract_contents(arg, global_scope) for arg in node.args
            )
            kw_args = tuple(
                f"{kw.arg}={self._extract_contents(kw.value, global_scope)}"
                for kw in node.keywords
            )
            if isinstance(node.func, ast.Name):
                return f"{node.func.id}{pos_args+kw_args}"
            elif isinstance(node.func, ast.Attribute):
                return f"{node.func.attr}{pos_args+kw_args}"
        elif isinstance(node, ast.Constant):
            return node.s
        else:
            return None

    def _extract_setup_parameters(self, tree, global_scope):
        """
        Returns a dictionary containing the parameters in the `setup` function,
        by extracting them from the given syntax tree.
        """
        for node in ast.walk(tree):
            if (
                isinstance(node, ast.Call)
                and isinstance(node.func, ast.Name)
                and node.func.id == "setup"
            ):
                parameters = {}
                for keyword in node.keywords:
                    parameter_name = keyword.arg
                    parameter_value = self._extract_contents(
                        keyword.value, global_scope
                    )
                    parameters[parameter_name] = parameter_value
                return parameters

        return None

    def _populate_global_scope(self, tree):
        """
        Returns a dictionary containing variables defined globally in the given syntax tree.
        """
        global_scope = {}
        for node in tree.body:
            if (
                isinstance(node, ast.Assign)
                and len(node.targets) == 1
                and isinstance(node.targets[0], ast.Name)
            ):
                target_name = node.targets[0].id
                target_value = self._extract_contents(node.value, global_scope)
                global_scope[target_name] = target_value
        return global_scope

    def _parse_entry_point(self, ep):
        """
        Example of expected entrypoint format:
        analytic_dynamics = cannon_analytic.analytic_dynamics:main
        """
        try:
            node_name = ep.split("=")[0].strip()
            entry_point = ep.split("=")[1].strip()
            file_name = f"{entry_point.split(':')[0].split('.')[1]}.py"
        except Exception as e:
            self.log.error(f"Failed to parse entry point {ep}")
            return None, None, None

        return node_name, entry_point, file_name

    def parse_setup_py(self, package_path):
        """
        Parse a setup.py file in the given package path.

        :param package_path: Path of the package.
        """
        package_path = Path(package_path)
        path_to_setup = str(Path(package_path, "setup.py"))

        try:
            with open(path_to_setup, "r") as f:
                source = f.read()
        except FileNotFoundError as e:
            self.log.error(f"Failed to find {path_to_setup}")
            return {}

        tree = ast.parse(source)
        global_scope = self._populate_global_scope(tree)
        parameters = self._extract_setup_parameters(tree, global_scope)

        if not parameters:
            self.log.error("Failed to parse setup.py")
            return {}

        nodes = []

        if (
            not parameters["entry_points"]
            or not parameters["entry_points"]["console_scripts"]
        ):
            self.log.warning(
                f"No entry points have been defined in `setup.py` of package {package_path}"
            )
        else:
            for console_script in parameters["entry_points"]["console_scripts"]:
                node_name, entry_point, file_name = self._parse_entry_point(
                    console_script
                )

                # print(f"    node: {node_name}")

                nodes.append(
                    {
                        "name": node_name,
                        "entry_point": entry_point,
                        "path": str(Path(package_path, package_path.name, file_name)),
                        "parameters": [],
                    }
                )

        return {
            "setup_py": path_to_setup,
            "package_name": package_path.name,
            "version": parameters.get("version", ""),
            "maintainer": parameters.get("maintainer", ""),
            "maintainer_email": parameters.get("maintainer_email", ""),
            "description": parameters.get("description", ""),
            "nodes": nodes,
        }

    ################################# Project #################################

    def get_project_packages(self, project_path, package_paths):
        """
        Collects packages metadata (nodes, parameters etc.)
        """
        self.log.debug(f" + get_project_packages {package_paths}")

        packages = []
        for package_path in package_paths:
            self.log.debug(f"package_path: {package_path}")
            # print(f"parsing package: {package_path}")

            parsed_data = None
            try:
                parsed_data = self.parse_xml(package_path)
            except ET.ParseError as e:
                print(
                    f"The path `{package_path}` doesn't contain xml, probably not a package. Skipping."
                )
                continue

            if parsed_data["build_type"] == "ament_python":
                temp = self.parse_setup_py(package_path)
                parsed_data["nodes"] = temp["nodes"]
                parsed_data["setup_py"] = temp["setup_py"]

            elif parsed_data["build_type"] == "ament_cmake":
                temp = self.parse_makefile(package_path)
                parsed_data["nodes"] = temp["nodes"]
                parsed_data["cmake"] = temp["cmake"]

            else:
                msg = f"Build type {parsed_data['build_type']} not allowed for ROS 2. Skipping `{package_path}`."
                self.log.error(msg)
                print(msg, color="yellow")
                continue

            all_parameters = {}
            package_parameters = {}

            path_to_config = os.path.join(package_path, "config", "params.yaml")
            if os.path.exists(path_to_config):
                with open(path_to_config, "r") as config_file:
                    try:
                        config = yaml.safe_load(config_file)
                    except yaml.YAMLError as ye:
                        self.log.exception(
                            f"Error parsing params.yaml in package {package_path}: {ye}"
                        )
                        raise ye

                for node_name, val in config.items():
                    par_dict = val.get("ros__parameters", {})
                    all_parameters[node_name] = []
                    for key, val in par_dict.items():
                        all_parameters[node_name].append(
                            {
                                "name": key,
                                "parameterType": type(val).__name__,
                                "value": val,
                                "description": "Parameter loaded from config.yaml",
                            }
                        )
                # first, populate parameters that belong to actual nodes.
                nodes_with_params = []
                for node in parsed_data["nodes"]:
                    node["parameters"] = all_parameters.get(node["name"], [])
                    nodes_with_params.append(node["name"])

                # then, populate package parameters, i.e. those that don't belong to any actual nodes.
                for name, param_list in all_parameters.items():
                    if not any(name == node_name for node_name in nodes_with_params):
                        package_parameters[name] = param_list

            else:
                self.log.info(f"no params.yaml file for package {package_path}")

            package_launch_paths = glob.glob(
                os.path.join(package_path, "**", "*.launch.py"), recursive=True
            )

            package_launch_paths = [
                plp for plp in package_launch_paths if plp not in self.ignore_list
            ]

            packages.append(
                {
                    "name": parsed_data["package_name"],
                    "cover": "",
                    "path": package_path,
                    "setup_py": parsed_data.get("setup_py", ""),
                    "package_xml": parsed_data.get("package_xml", ""),
                    "maintainer": parsed_data.get("maintainer", ""),
                    "maintainer_email": parsed_data.get("maintainer_email", ""),
                    "description": parsed_data.get("description", ""),
                    "git": "",
                    "launches": self.get_project_launch_files(package_launch_paths),
                    "nodes": parsed_data.get("nodes", []),
                    "parameters": package_parameters,
                }
            )

        return packages

    def get_project_launch_files(self, launch_paths):
        """
        returns a list of dictionaries describing the *.launch.py files in the given list.
        """
        launch_files = []
        for launch_path in launch_paths:
            launch_files.append(
                {
                    "name": Path(launch_path).name,
                    "path": launch_path,
                    # "tags": [],
                    "description": "",
                }
            )

        launch_files.sort(key=lambda x: x["name"])
        return launch_files

    def get_git_remote_url(self, project_path, no_remote_ok=False):
        """
        Get the URL of the 'origin' remote.

        Assumption: there exists a remote named 'origin'.
        """
        try:
            repo = git.Repo(project_path)
            default_remote_name = "origin"
            if default_remote_name in repo.remotes:
                default_remote = repo.remotes[default_remote_name]
                default_remote_url = default_remote.config_reader.get("url")

                # sanity check
                if default_remote_url.endswith(".git"):
                    return default_remote_url
                else:
                    self.log.error(
                        "Could not obtain git remote url for path " + project_path
                    )
                    return ""
            else:
                msg = f"{project_path} has no git remote named `origin`."
                if no_remote_ok:
                    # self.log.info(msg)
                    pass
                else:
                    self.log.error(msg)
                return ""

        except Exception as e:
            self.log.exception(e)
            return ""

    def get_git_local_hash(self, project_path):
        try:
            # Get the hash of the latest commit on the current branch
            latest_commit_hash = git.Repo(project_path).head.commit.hexsha

        except Exception as e:
            self.log.exception(e)
            return None

        # sanity check
        if len(latest_commit_hash) == 40:
            return latest_commit_hash
        else:
            self.log.error("Could not obtain git hash for local path " + project_path)
            return None

    def get_project_description(self, project_path):        
        return ""

    def get_file_content(self, path):
        try:
            with open(path, "r") as f:
                content = f.read()
            return content
        except Exception as e:
            self.log.error(f"could not open file {path}")
            return ""

    def _setup_project(self, project_path):
        from citros_meta import __version__

        self.project = {
            "citros_cli_version": __version__,
            "cover": "",
            "tags": [],
            "is_active": True,
            "description": "",
            "git": self.get_git_remote_url(project_path),
            # "path": project_path,
            "packages": None,
            "launches": None,
        }

    def _is_descendant(self, file_path, directory_path):
        file = Path(file_path).resolve()
        directory = Path(directory_path).resolve()
        return str(directory) in str(file)

    def get_project_package_paths(self, project_path: str):
        pkg_xmls = glob.glob(
            os.path.join(project_path, "**", "src", "**", "package.xml"), recursive=True
        )
        pkg_xmls.sort()

        # get the directories containing the package.xml files
        package_paths = [os.path.dirname(file) for file in pkg_xmls]

        # save paths relative to project path
        package_paths = [os.path.relpath(pp, project_path) for pp in package_paths]

        # filter files from ignore list
        package_paths = [pp for pp in package_paths if pp not in self.ignore_list]

        return package_paths

    def get_project_launch_paths(self, project_path, package_paths):
        # project-level launch files are either descendants of src but not inside a package,
        # or directly under project_path/launch
        launch_paths = glob.glob(
            os.path.join(project_path, "**", "src", "**", "*.launch.py"), recursive=True
        )
        project_launch_paths = [
            lp
            for lp in launch_paths
            if not any(self._is_descendant(lp, pkg) for pkg in package_paths)
        ]
        project_launch_paths = project_launch_paths + glob.glob(
            os.path.join(project_path, "launch", "*.launch.py")
        )

        # filter ignored files
        project_launch_paths = [
            plp for plp in project_launch_paths if plp not in self.ignore_list
        ]

        return project_launch_paths

    def parse(self, project_path: str):
        """
        parses the project with the given name under the given path.

        Assumptions:

            1. a package is a directory that holds a `package.xml` file,
               and is a decendant of a `src` directory,
               which is a descendant of the project directory.

            2. a launch file is a file whose name ends with `.launch.py`,
               and is a decendant of a `src` directory,
               which is a descendant of the project directory.

        param: project_path - the project directory's full path


        Returns:
        a dictionary holding all metadata for the project.
        """
        packages = []
        launches = []

        if not self.project:
            self._setup_project(project_path)

        package_paths = self.get_project_package_paths(project_path)

        project_launch_paths = self.get_project_launch_paths(
            project_path, package_paths
        )

        packages = self.get_project_packages(project_path, package_paths)
        launches = self.get_project_launch_files(project_launch_paths)

        self.project["description"] = self.get_project_description(project_path)
        self.project["packages"] = packages
        self.project["launches"] = launches

        return self.project

    def generate_default_proj_json(self, proj_name, package_names=[]):
        proj_json = {"name": f"{proj_name}", "packages": []}

        for pkg in package_names:
            pkg_content = {
                "name": f"{pkg}",
                "nodes": [],
                "launches": [],
                "parameters": [],
            }
            proj_json["packages"].append(pkg_content)

        return proj_json

    def get_msg_files(self, project_path, from_install=True):
        if from_install:
            msg_paths = glob.glob(
                os.path.join("**", "install", "**", "share", "**", "msg", "*.msg"),
                recursive=True,
            )
            return msg_paths
        else:
            msg_paths = glob.glob(
                os.path.join(project_path, "**", "*.msg"), recursive=True
            )
            return msg_paths

    def generate_default_params_setup(self, citros, override_existing=False):
        packages = citros["packages"]

        json_data = {"packages": {}}

        for package in packages:
            package_dict = {package["name"]: {}}
            for node in package["nodes"]:
                node_dict = {node["name"]: {"ros__parameters": {}}}
                for parameter in node["parameters"]:
                    node_dict[node["name"]]["ros__parameters"][
                        parameter["name"]
                    ] = parameter["value"]
                package_dict[package["name"]][node["name"]] = node_dict[node["name"]]

            # node parameters
            json_data["packages"][package["name"]] = package_dict[package["name"]]

            # package parameters
            for node_name, param_list in package["parameters"].items():
                node_dict = {node_name: {"ros__parameters": {}}}
                for parameter in param_list:
                    node_dict[node_name]["ros__parameters"][
                        parameter["name"]
                    ] = parameter["value"]
                json_data["packages"][package["name"]] = node_dict
        return json_data
        # with open(params_setup_json, "w") as f:
        #     json.dump(json_data, f, indent=4)

    ####################### verify user defined functions #######################

    def find_function_in_file(self, file_path, function_name):
        """Scan a Python file to check if a global function exists with the given function_name."""
        try:
            self.log.debug(f"opening {file_path} to look for {function_name}.")
            with open(file_path, "r") as file:
                tree = ast.parse(file.read())
                for node in ast.walk(tree):
                    if isinstance(node, ast.FunctionDef) and node.name == function_name:
                        return True
        except Exception as e:
            msg = f"exception was raise while trying to find function `{function_name}` in file {file_path}."
            self.log.error(f"{msg}\n{e}")
            print(f"{msg}", color="red")
            return False

        print(
            f"Could not find the user-defined function `{function_name}` in {file_path}.",
            color="red",
        )
        return False

    def scan_for_functions(self, data, functions_dir):
        """Recursively scan nested JSON data for the 'function' key."""
        if isinstance(data, dict):
            if "function" in data:
                value = data["function"]

                # Check if the value contains ':' (i.e. it's a user-defined function)
                if ":" in value:
                    file_name, func_name = value.split(":")
                    if not self.find_function_in_file(
                        Path(functions_dir, file_name), func_name
                    ):
                        # Found the key-value pair, but function doesn't exist
                        return False

            for key in data:
                if not self.scan_for_functions(data[key], functions_dir):
                    return False

        # unnecessary, but just in case...
        elif isinstance(data, list):
            for item in data:
                if not self.scan_for_functions(item, functions_dir):
                    return False

        return True

    def check_user_defined_functions(self, param_setup_dir: Path):
        for param_setup_file in param_setup_dir.iterdir():
            if param_setup_file.is_file() and str(param_setup_file).endswith(".json"):
                self.log.debug(
                    f"checking user-defined functions for {param_setup_file}."
                )
                with open(param_setup_file, "r") as file:
                    data = json.load(file)
                    if not self.scan_for_functions(
                        data, Path(param_setup_dir, "functions")
                    ):
                        return False
        return True
